"""
The ``kszx.sdss`` module contains functions for downloading/parsing SDSS data products.

References:
   - https://data.sdss.org/sas/dr12/boss/lss/
   - https://data.sdss.org/datamodel/files/BOSS_LSS_REDUX/
"""

import os
import fitsio

from . import Catalog
from . import io_utils


def read_galaxies(survey, dr=12, download=False):
    r"""Reads SDSS galaxy catalog, and returns a Catalog object.
    
    After calling this function, you'll probably want to call :func:`Catalog.apply_redshift_cut()`
    to retrict to an appropriate redshift range.

    Function arguments:

      - ``survey`` (str): either 'CMASS_North', 'CMASS_South', 'LOWZ_North', 'LOWZ_South',
        'CMASSLOWZTOT_North', 'CMASSLOWZTOT_South', 'CLASSLOWZE2_North', or 'CMASSLOWZE3_North'.

      - ``dr`` (integer): either 11 or 12.

      - ``download`` (boolean): if True, then all needed data files will be auto-downloaded.

    Returns a :class:`kszx.Catalog` object, with the following columns::
     
     ra_deg, dec_deg, z,    # sky location, redshift
     wfkp,                  # FKP weight
     wcp, wzf, wsys,        # systematic weights
     cboss,                 # completeness (denoted C_BOSS in papers and COMP in the FITS file)
     id                     # unique identifier

    Reminder: SDSS galaxies are usually weighted by ``(wzf + wcp − 1) * wsys * wkp``.

    Example usage::

     # kszx.sdss.read_galaxies() returns a kszx.Catalog.
     gcat = kszx.sdss.read_galaxies('CMASS_North')
     gcat.apply_redshift_cut(0.43, 0.7)
    """
    
    filename = _galaxy_filename(survey, dr, download, dlfunc='kszx.sdss.read_galaxies')
    gcat = read_fits_catalog(filename, is_randcat=False)
    print("Reminder: you probably want to call cat.apply_redshift_cut(zmin,zmax)"
          + " on the Catalog returned by kszx.sdss.read_galaxies())")
    return gcat


def read_randoms(survey, dr=12, download=False):
    r"""Reads SDSS random catalog, and returns a Catalog object.

    After calling this function, you'll probably want to call :func:`Catalog.apply_redshift_cut()`
    to retrict to an appropriate redshift range.

    Function arguments:

      - ``survey`` (str): either 'CMASS_North', 'CMASS_South', 'LOWZ_North', 'LOWZ_South',
        'CMASSLOWZTOT_North', 'CMASSLOWZTOT_South', 'CLASSLOWZE2_North', or 'CMASSLOWZE3_North'.

      - ``dr`` (integer): either 11 or 12.

      - ``download`` (boolean): if True, then all needed data files will be auto-downloaded.

    Returns a :class:`kszx.Catalog` object, with the following columns::
     
     ra_deg, dec_deg, z   # sky location, redshift
     wfkp                 # FKP weight

    Example usage::

     # kszx.sdss.read_randoms() returns a kszx.Catalog.
     rcat = kszx.sdss.read_randoms('CMASS_North')
     rcat.apply_redshift_cut(0.43, 0.7)
    """
    
    filenames = _random_filenames(survey, dr, download, dlfunc='kszx.sdss.read_randoms')
    catalogs = [ read_fits_catalog(f, is_randcat=True) for f in filenames ]
    rcat = Catalog.concatenate(catalogs, name=f'{survey} randoms', destructive=True)
    print("Reminder: you probably want to call cat.apply_redshift_cut(zmin,zmax)"
          + " on the Catalog returned by kszx.sdss.read_randoms())")
    return rcat


def read_mask(survey, dr=12, download=False):
    r"""Reads SDSS mangle mask, and returns a pymangle object. Rarely needed!

    Function arguments:

      - ``survey`` (str): either 'CMASS_North', 'CMASS_South', 'LOWZ_North', 'LOWZ_South',
        'CMASSLOWZTOT_North', 'CMASSLOWZTOT_South', 'CLASSLOWZE2_North', or 'CMASSLOWZE3_North'.

      - ``dr`` (integer): either 11 or 12.

      - ``download`` (boolean): if True, then all needed data files will be auto-downloaded.

    Returns a pymangle object that can be evaluated with mask.weight(ra_deg, dec_deg).
    The return value is either 0, or close to 1 (e.g. 0.98).

    Note: requires pymangle! (Not a dependency by default.)    
    """
    import pymangle
    filename = _mask_filename(survey, dr, download, dlfunc='kszx.sdss.read_mask')
    print(f'Reading {filename}\n', end='')
    return pymangle.Mangle(filename)


def read_mock(survey, mock_type, ix, dr=12, download=False):
    r"""Reads SDSS mock galaxy catalog, and returns a Catalog object.

    Note that the qpm mocks are already restricted to an appropriate redshift range,
    i.e. there should be no need to call :func:`Catalog.apply_redshift_cut()` (unlike
    the "core" SDSS galaxy/random catalogs).

    Function arguments:

      - ``survey`` (str): either 'CMASS_North', 'CMASS_South', 'LOWZ_North', 'LOWZ_South'.

      - ``mock_type`` (str): currently 'qpm' (DR12 only) and 'pthalos' (DR11 only)
        are implemented.

      - ``ix`` (integer): index of mock, in the range ``0 <= ix < 1000``, except for
        CMASS PTHALOS where the range is ``0 <= ix < 600``.

      - ``dr`` (integer): either 11 or 12.

      - ``download`` (boolean): if True, then all needed data files will be auto-downloaded.

    Returns a :class:`kszx.Catalog` object, with the following columns::
     
     ra_deg, dec_deg, z,    # sky location, redshift
     wveto                  # systematic weights
     wfkp                   # FKP weight (available for QPM but not PTHALOS)
     cboss, wcp, wzf        # PTHALOS only: completeness/systematics weights
     ztrue                  # PTHALOS only: true redshift, removing peculiar velocities

    Example usage::

     # kszx.sdss.read_mock() returns a kszx.Catalog.
     mcat = kszx.sdss.read_mock('CMASS_North', 'qpm', 0)   # index can be 0 <= ix < 1000
     # mcat.apply_redshift_cut(0.43, 0.7) not needed!
    """

    filename = _mock_filename(survey, mock_type, ix, dr, download=download, dlfunc='kszx.sdss.read_mock')

    if mock_type.lower() == 'qpm':
        # https://data.sdss.org/datamodel/files/BOSS_LSS_REDUX/dr11_qpm_mocks/mock_galaxy_DRX_SAMPLE_NS_QPM_IDNUMBER.html
        return Catalog.from_text_file(filename, col_names=['ra_deg','dec_deg','z','wfkp','wveto'])
    elif mock_type.lower() == 'pthalos':
        # https://data.sdss.org/datamodel/files/BOSS_LSS_REDUX/dr11_pthalos_mocks/mock_galaxy_DRX_SAMPLE_NS_PTHALOS_IDNUMBER.html
        # Omit 'ipoly' and 'galaxy flag'
        return Catalog.from_text_file(filename, col_names=['ra_deg','dec_deg','z', None,'cboss','wcp','wzf','wveto','ztrue',None])
    else:
        raise RuntimeError(f"kszx: SDSS {mock_type=} was not recognized (currently support 'qpm' and 'pthalos')")


def read_mock_randoms(survey, mock_type, dr=12, download=False):
    r"""Reads SDSS mock random catalog, and returns a Catalog object.

    Note that the qpm mocks are already restricted to an appropriate redshift range,
    i.e. there should be no need to call :func:`Catalog.apply_redshift_cut()` (unlike
    the "core" SDSS galaxy/random catalogs).

    Function arguments:

      - ``survey`` (str): either 'CMASS_North', 'CMASS_South', 'LOWZ_North', 'LOWZ_South'.

      - ``mock_type`` (str): currently 'qpm' (DR12 only) and 'pthalos' (DR11 only)
        are implemented.

      - ``dr`` (integer): either 11 or 12.

      - ``download`` (boolean): if True, then all needed data files will be auto-downloaded.

    Returns a :class:`kszx.Catalog` object, with the following columns::
     
     ra_deg, dec_deg, z   # sky location, redshift
     wfkp                 # FKP weight

    Example usage::

     # kszx.sdss.read_mock_randoms() returns a kszx.Catalog.
     mrcat = kszx.sdss.read_mock_randoms('CMASS_North', 'qpm')
     # mrcat.apply_redshift_cut(0.43, 0.7) not needed!
    """

    filenames = _mock_random_filenames(survey, mock_type, dr, download=download, dlfunc='kszx.sdss.read_mock_randoms')

    if mock_type.lower() == 'qpm':
        # https://data.sdss.org/datamodel/files/BOSS_LSS_REDUX/dr11_qpm_mocks/mock_random_DRX_SAMPLE_NS_QPM_NxV.html
        catalogs = [ Catalog.from_text_file(f, col_names=['ra_deg','dec_deg','z','wfkp']) for f in filenames ]
        return Catalog.concatenate(catalogs, name=f'{survey} {mock_type} mock randoms', destructive=True)
    elif mock_type.lower() == 'pthalos':
        # https://data.sdss.org/datamodel/files/BOSS_LSS_REDUX/dr11_pthalos_mocks/mock_random_DRX_SAMPLE_NS_PTHALOS_IDNUMBER.html
        catalogs = [ Catalog.from_text_file(f, col_names=['ra_deg','dec_deg','z',None,'cboss','wcp','wzf','wveto']) for f in filenames ]
        return Catalog.concatenate(catalogs, name=f'{survey} {mock_type} randoms', destructive=True)        
    else:
        raise RuntimeError(f"kszx: SDSS {mock_type=} was not recognized (currently support 'qpm' and 'pthalos')")
    

def download(survey, dr=12, mask=False, qpm=False, pthalos=False):
    r"""Downloads SDSS data products (galaxies, randoms, and optionally mocks + mangle mask) for a given survey.

    Can be called from command line: ``python -m kszx download_sdss``.

    Function arguments:

      - ``survey`` (str): either 'CMASS_North', 'CMASS_South', 'LOWZ_North', 'LOWZ_South',
        'CMASSLOWZTOT_North', 'CMASSLOWZTOT_South', 'CLASSLOWZE2_North', or 'CMASSLOWZE3_North'.

      - ``dr`` (integer): either 11 or 12.

      - ``mask`` (boolean): if True, then mangle mask will be downloaded (in addition to 
        galaxies and randoms).

      - ``qpm`` (boolean): if True, then QPM mocks will be downloaded (DR12 only).

      - ``pthalos`` (boolean): if True, then PTHALOS mocks will be downloaded (DR11 only).
    """
    
    _galaxy_filename(survey, dr, download=True)
    _random_filenames(survey, dr, download=True)
    _mask_filename(survey, dr, download=mask)

    for flag, mock_type in [ (qpm,'qpm'), (pthalos,'pthalos') ]:
        if not flag:
            continue
        for _ in _mock_random_filenames(survey, mock_type, dr, download=True):
            pass
        _mock_filename(survey, mock_type, 0, dr, download=True)  # will download all mocks


####################################################################################################


def read_fits_catalog(filename, is_randcat, name=None, extra_columns=[]):
    r"""Reads FITS file in SDSS catalog format. 

    Intended as a helper for read_galaxies() or read_randoms(), but may be useful elsewhere.
    
    Function arguments:

      - ``filename`` (string): should end in ``.fits``.
      - ``is_randcat`` (boolean): True for a random catalog, False for a galaxy catalog.
        (Determines which column names are expected).
      - ``name`` (str, optional): name of Catalog, passed to :class:`kszx.Catalog` constructor.
      - ``extra_columns`` (list of string): extra columns to read (if any).

    Returns a :class:`kszx.Catalog` object.
    """

    print(f'Reading {filename}\n', end='')
    catalog = Catalog(name=name, filename=filename)

    with fitsio.FITS(filename) as f:
        
        # For a list of all fields, see
        #   https://data.sdss.org/datamodel/files/BOSS_LSS_REDUX/galaxy_DRX_SAMPLE_NS.html
        #   https://data.sdss.org/datamodel/files/BOSS_LSS_REDUX/randomN_DRX_SAMPLE_NS.html
        
        catalog.add_column('ra_deg', f[1].read('RA'))
        catalog.add_column('dec_deg', f[1].read('DEC'))
        catalog.add_column('z', f[1].read('Z'))
        catalog.add_column('wfkp', f[1].read('WEIGHT_FKP'))

        # catalog.add_column('ipoly', f[1].read('IPOLY'))
        # catalog.add_column('isect', f[1].read('ISECT'))

        if not is_randcat:
            catalog.add_column('wcp', f[1].read('WEIGHT_CP'))
            catalog.add_column('wzf', f[1].read('WEIGHT_NOZ'))
            catalog.add_column('wsys', f[1].read('WEIGHT_SYSTOT'))
            catalog.add_column('cboss', f[1].read('COMP'))
            catalog.add_column('id', f[1].read('ID'))

        for col_name in extra_columns:
            catalog.add_column(col_name.lower(), f[1].read(col_name.upper()))

    catalog._announce_file_read()
    return catalog


####################################################################################################


def _dr_str(dr):
    dhash = { 11: 'DR11v1', 12: 'DR12v5' }
    
    if dr not in dhash:
        raise RuntimeError(f"SDSS {dr=} not supported (currently support DR11, DR12)")

    return dhash[dr]


def _check_survey(survey, dr, abridged=False):
    """Checks 'survey' for validity, and returns its standard capitalization."""

    if dr not in [11,12]:
        raise RuntimeError(f"SDSS {dr=} not supported (currently support DR11, DR12)")
    
    survey_list = [ 'CMASS_North', 'CMASS_South', 'LOWZ_North', 'LOWZ_South' ]

    if (dr==12) and (not abridged):
        survey_list += [ 'CMASSLOWZTOT_North', 'CMASSLOWZTOT_South',
                         'CLASSLOWZE2_North', 'CMASSLOWZE3_North' ]
                    
    for s in survey_list:
        if survey.upper() == s.upper():
            return s

    raise RuntimeError(f"SDSS survey '{survey}' not recognized (must be one of: {survey_list})")
        
    
def _sdss_path(relpath, dr, download, *, packed_relpath=None, gz=False, dlfunc=None):
    """Intended to be called through wrapper such as _galaxy_filename(), _random_filenames(), etc.

    If 'packed_relpath' is specified, it should be a file ending in '.gz', '.tgz', or '.tar.gz'
    which generates 'relpath' after unpacking.

    Setting gz=True is shorthand for packed_relpath = "{relpath}.gz".

    The 'dr' argument is only used if download=True, to construct the URL.
    
    Here and in other parts of kszx, the 'dlfunc' argument gives the name of a transitive caller that
    expects the file to be present, and has a 'download=False' optional argument. This information is
    only used when generating exception-text (to tell the user how to download the file).

    Example: 

         _sdss_path(relpath = 'qpm_mocks/mock_galaxy_DR12_LOWZ_S_QPM_0137.rdzw',
                    packed_relpath = 'qpm_mocks/mock_galaxy_DR12_LOWZ_S_QPM_allmocks.tar.gz',
                    download = True, dr = 12)

    Does the following:

       - If the following file is found, then return:
          {kszx_data_dir}/sdss/DR12v5/qpm_mocks/mock_galaxy_DR12_LOWZ_S_QPM_0137.rdzw

       - If the following file is found, then unpack and return:
          {kszx_data_dir}/sdss/DR12v5/qpm_mocks/mock_galaxy_DR12_LOWZ_S_QPM_allmocks.tar.gz

       - Download and decompress:
           https://data.sdss.org/sas/dr12/boss/lss/qpm_mocks/mock_galaxy_DR12_LOWZ_S_QPM_allmocks.tar.gz
    """

    if gz and (packed_relpath is not None):
        raise RuntimeError("kszx.sdss internal error: can't specify both 'packed_relpath' and 'gz'")
    
    sdss_base_dir = os.path.join(io_utils.get_data_dir(), 'sdss')
    abspath = os.path.join(sdss_base_dir, _dr_str(dr), relpath)

    if not io_utils.do_download(abspath, download, dlfunc):
        return abspath

    if gz:
        packed_relpath = relpath + '.gz'

    if not packed_relpath:
        # Case 1: No tar/gz involved, just download file.
        url = f'https://data.sdss.org/sas/dr{dr}/boss/lss/{relpath}'
        io_utils.wget(abspath, url)   # calls assert os.path.exists(abspath) after downloading
    else:
        # Case 2: Download tar/gz file by calling _sdss_path() recursively, then unpack.
        packed_abspath = _sdss_path(packed_relpath, dr, download=True)
        io_utils.unpack(packed_abspath, abspath)  # calls assert os.path.exists(abspath) after unpacking
    
    return abspath


def _galaxy_filename(survey, dr, download=False, dlfunc=None):
    d = _dr_str(dr)
    s = _check_survey(survey, dr)
    return _sdss_path(f'galaxy_{d}_{s}.fits', dr, download, gz=True, dlfunc=dlfunc)


def _random_filenames(survey, dr, download=False, dlfunc=None):
    d = _dr_str(dr)
    s = _check_survey(survey, dr)
    return [ _sdss_path(f'random{n}_{d}_{s}.fits', dr, download, gz=True, dlfunc=dlfunc) for n in [0,1] ]


def _mask_filename(survey, dr, download=False, dlfunc=None):
    d = _dr_str(dr)
    s = _check_survey(survey, dr)
    return _sdss_path(f'mask_{d}_{s}.ply', dr, download, dlfunc=dlfunc)


def _mock_filename(survey, mock_type, ix, dr, download=False, dlfunc=None):
    _check_survey(survey, dr, abridged=True)

    if (dr == 12) and (mock_type.lower() == 'qpm'):
        assert 0 <= ix < 1000
        
        # The 'ix' argument is a zero-based index 0 <= ix < 1000, but the
        # QPM files use a one-based index 1 <= ix <= 1000, so we add 1.
            
        return _sdss_path(
            relpath = f'qpm_mocks/mock_galaxy_DR{dr}_{survey[:-4].upper()}_QPM_{(ix+1):04d}.rdzw',
            packed_relpath = f'qpm_mocks/mock_galaxy_DR{dr}_{survey[:-4].upper()}_QPM_allmocks.tar.gz',
            download = download,
            dr = dr,
            dlfunc = dlfunc
        )

    elif (dr == 11) and (mock_type.lower() == 'pthalos'):
        assert 0 <= ix < 600
        
        # The 'ix' argument is a zero-based index 0 <= ix < 600, but the
        # pthalos mocks use an one-index 4001 <= ir <= 4600, so we add 4001 (!)

        return _sdss_path(
            relpath = f'dr{dr}_pthalos_mocks/mock_galaxy_DR{dr}_{survey[:-4].upper()}_PTHALOS_ir{(ix+4001)}.dat',
            packed_relpath = f'dr{dr}_pthalos_mocks/mock_galaxy_DR{dr}_{survey[:-4].upper()}_PTHALOS_allmocks.tar.gz',
            download = download,
            dr = dr,
            dlfunc = dlfunc
        )
    
    else:
        raise RuntimeError(f"kszx: SDSS {(mock_type,dr)=} was not recognized (currently support ('qpm',12) and ('pthalos',11))")


def _mock_random_filenames(survey, mock_type, dr, download=False, dlfunc=None):
    _check_survey(survey, dr, abridged=True)
    
    if (dr == 12) and (mock_type.lower() == 'qpm'):
        for n in [1,2]:
            yield _sdss_path(f'qpm_mocks/mock_random_DR{dr}_{survey[:-4].upper()}_50x{n}.rdzw', dr=dr, download=download, gz=True, dlfunc=dlfunc)

    elif (dr == 11) and (mock_type.lower() == 'pthalos'):
        nfiles = 600 if survey.lower().startswith('cmass') else 1000
        for n in range(4001, 4001+nfiles):
            yield _sdss_path(
                relpath = f'dr{dr}_pthalos_mocks/mock_random_DR{dr}_{survey[:-4].upper()}_PTHALOS_ir{n}.dat',
                packed_relpath = f'dr{dr}_pthalos_mocks/mock_random_DR{dr}_{survey[:-4].upper()}_PTHALOS_allmocks.tar.gz',
                download = download,
                dr = dr,
                dlfunc = dlfunc
            )
    
    else:
        raise RuntimeError(f"kszx: SDSS {(mock_type,dr)=} was not recognized (currently support ('qpm',12) and ('pthalos',11))")



import os
import healpy
import scipy.fft
import pixell.enmap
import numpy as np

from .Box import Box
from . import cpp_kernels
from . import utils
    

####################################################################################################


def fft_r2c(box, arr, spin=0, threads=None):
    r"""Computes the FFT of real-space map 'arr', and returns a Fourier-space map.

        - ``box`` (kszx.Box): defines pixel size, bounding box size, and location of observer.
          See :class:`~kszx.Box` for more info.

        - ``arr`` (array): numpy array representing a real-space map (dtype=float).

        - ``spin`` (integer): currently only spin=0 and spin=1 are supported.

        - ``threads`` (integer or None): number of parallel threads used.
          If ``threads=None``, then number of threads defaults to :func:`~kszx.utils.get_nthreads()`.

    Returns a numpy array representing a Fourier-space map (dtype=complex).

    The real-space and Fourier-space array shapes are given by ``box.real_space_shape``
    and ``box.fourier_space_shape``, and are related as follows:

    $$\begin{align}
    (\mbox{real-space shape}) &= (n_0, n_1, \cdots, n_{d-1}) \\
    (\mbox{Fourier-space shape}) &= (n_0, n_1, \cdots, \lfloor n_{d-1}/2 \rfloor + 1)
    \end{align}$$

    Notes:

       - Our spin-0 Fourier conventions are:

          $$f(k) = V_{pix} \sum_x f(x) e^{-ik\cdot x}$$

          $$f(x) = V_{box}^{-1} \sum_k f(k) e^{ik\cdot x}$$

       - We define spin-1 Fourier transforms by inserting an extra factor
         $(\pm i {\hat k} \cdot {\hat r})$:

          $$f(k) = V_{pix} \sum_x f(x) (-i {\hat k} \cdot {\hat r}) e^{-ik\cdot x}$$

          $$f(x) = V_{box}^{-1} \sum_k f(k) (i {\hat k} \cdot {\hat r}) e^{ik\cdot x}$$

         where the line-of-sight direction $\hat r$ is defined in "observer coordinates"
         (see :class:`~kszx.Box` for more info).

         Application: the spin-1 c2r transform can be used to compute the radial velocity
         field from the density field (with a factor $faH/k$).

         Another application: the spin-1 r2c transform can be used to estimate $P_{gv}(k)$
         or $P_{vv}(k)$ from the radial velocity field (or the kSZ velocity reconstruction),
         by calling :func:`~kszx.fft_r2c()` followed by :func:`~kszx.estimate_power_spectrum()`.

       - At some point in the future, I'll define spin-$l$ transforms, with an
         extra factor $(\pm i^l P_l({\hat k} \cdot {\hat r}))$.
    """

    assert isinstance(box, Box)
    assert box.is_real_space_map(arr)   # check shape and dtype of input array

    if threads is None:
        threads = utils.get_nthreads()
    
    if spin == 0:
        # Currently using scipy.fft instead of pyfftw.
        # (When I tried pyfftw, it was weirdly slow.)
        ret = scipy.fft.rfftn(arr, workers=threads)
        ret *= box.pixel_volume   # see Fourier conventions in docstring
        return ret                # numpy array with shape=box.fourier_space_shape and dtype=complex.

    if spin != 1:
        raise RuntimeError('fft_r2c(): currently we support only spin=0 and spin=1')

    # Spin-1 FFT follows.
    
    ret = np.zeros(box.fourier_space_shape, dtype=complex)
    
    for axis in range(box.ndim):
        t = multiply_rfunc(box, arr, lambda r: 1./r, regulate=True)
        t *= box.get_r_component(axis)
        t = fft_r2c(box, t, spin=0, threads=threads)
        t *= -1j * box.get_k_component(axis, zero_nyquist=True)  # note minus sign here
        ret += t
        del t

    multiply_kfunc(box, ret, lambda k: 1./k, dc=0, in_place=True)
    return ret
    

def fft_c2r(box, arr, spin=0, threads=None):
    r"""Computes the FFT of Fourier-space map 'arr', and returns a real-space map.

        - ``box`` (kszx.Box): defines pixel size, bounding box size, and location of observer.
          See :class:`~kszx.Box` for more info.

        - ``arr``: numpy array representing a Fourier-space map (dtype=complex).

        - ``spin``: currently only spin=0 and spin=1 are supported.

        - ``threads`` (integer or None): number of parallel threads used.
          If ``threads=None``, then number of threads defaults to :func:`~kszx.utils.get_nthreads()`.

    Returns a numpy array representing a real-space map (dtype=float).

    The real-space and Fourier-space array shapes are given by ``box.real_space_shape``
    and ``box.fourier_space_shape``, and are related as follows:

    $$\begin{align}
    (\mbox{real-space shape}) &= (n_0, n_1, \cdots, n_{d-1}) \\
    (\mbox{Fourier-space shape}) &= (n_0, n_1, \cdots, \lfloor n_{d-1}/2 \rfloor + 1)
    \end{align}$$

    Notes:

       - Our spin-0 Fourier conventions are (see Box docstring):

          $$f(k) = V_{pix} \sum_x f(x) e^{-ik\cdot x}$$

          $$f(x) = V_{box}^{-1} \sum_k f(k) e^{ik\cdot x}$$

       - We define spin-1 Fourier transforms by inserting an extra factor
         $(\pm i {\hat k} \cdot {\hat r})$:

          $$f(k) = V_{pix} \sum_x f(x) (-i {\hat k} \cdot {\hat r}) e^{-ik\cdot x}$$

          $$f(x) = V_{box}^{-1} \sum_k f(k) (i {\hat k} \cdot {\hat r}) e^{ik\cdot x}$$

         where the line-of-sight direction $\hat r$ is defined in "observer coordinates"
         (see :class:`~kszx.Box` for more info).

         Application: the spin-1 c2r transform can be used to compute the radial velocity
         field from the density field (with a factor $faH/k$).

         Another application: the spin-1 r2c transform can be used to estimate $P_{gv}(k)$
         or $P_{vv}(k)$ from the radial velocity field (or the kSZ velocity reconstruction),
         by calling :func:`~kszx.fft_r2c()` followed by :func:`~kszx.estimate_power_spectrum()`.

       - At some point in the future, I'll define spin-$l$ transforms, with an
         extra factor $(\pm i^l P_l({\hat k} \cdot {\hat r}))$.
    """
    
    assert isinstance(box, Box)
    assert box.is_fourier_space_map(arr)   # check shape and dtype of input array

    if threads is None:
        threads = utils.get_nthreads()

    if spin == 0:
        # Currently using scipy.fft instead of pyfftw.
        # (When I tried pyfftw, it was weirdly slow.)
        ret = scipy.fft.irfftn(arr, box.npix, workers=threads)
        ret *= (1.0 / box.pixel_volume)    # see Fourier conventions in docstring
        return ret                         # numpy array with shape=box.real_space shape and dtype=complex.

    if spin != 1:
        raise RuntimeError('fft_c2r(): currently we support only spin=0 and spin=1')

    # Spin-1 FFT follows.
    
    ret = np.zeros(box.real_space_shape, dtype=float)
    
    for axis in range(box.ndim):
        t = multiply_kfunc(box, arr, lambda k: 1./k, dc=0.)
        t *= 1j * box.get_k_component(axis, zero_nyquist=True)
        t = fft_c2r(box, t, spin=0)
        t *= box.get_r_component(axis)
        ret += t
        del t

    multiply_rfunc(box, ret, lambda r: 1./r, regulate=True, in_place=True)
    return ret


####################################################################################################


def interpolate_points(box, arr, points, kernel, fft=False, spin=0, periodic=False):
    r"""Interpolates real-space map at a specified set of points.

    Function args:

        - ``box`` (kszx.Box): defines pixel size, bounding box size, and location of observer.
          See :class:`~kszx.Box` for more info.

        - ``arr`` (numpy array): represents the map to be interppolated, either in real space 
          (if ``fft=False``) or Fourier space (if ``fft=True``).
          
          The real-space and Fourier-space array shapes are given by ``box.real_space_shape``
          and ``box.fourier_space_shape``, and are related as follows:
 
         $$\begin{align}
         (\mbox{real-space shape}) &= (n_0, n_1, \cdots, n_{d-1}) \\
         (\mbox{Fourier-space shape}) &= (n_0, n_1, \cdots, \lfloor n_{d-1}/2 \rfloor + 1)
         \end{align}$$

        - ``points`` (numpy array):
          Sequence of points where the map is to be interpolated.
          Array shape should be (n,d), where n is the number of interpolation points, and d
          is the box dimension (usually 3).

        - ``kernel`` (string): either ``'cic'`` or ``'cubic'`` (more options will be defined later).

        - ``fft`` (boolean): if True, then ``arr`` is a Fourier-space map, and ``fft_c2r(arr, spin)``
          will be called before interpolating.
    
        - ``spin`` (integer): passed as 'spin' argument to ``fft_c2r()``. (Only used if ``fft=True``.)

        - ``periodic`` (boolean): if True, then the box has periodic boundary conditions.

    Return value:

        - 1-d numpy array with length npoints, containing interpolated values.
 
    Notes: 

       - The ``points`` array should be specified in "observer coordinates", not "grid coordinates".

         (Reminder: in observer coordinates, the observer is at the origin, coordinates have units
         of distance, and the corners of the box are at ``box.{lpos,rpos}``. 
         See :class:`~kszx.Box` for more info.)

       - Before calling ``interpolate_points()``, you may want to call :func:`~kszx.apply_kernel_compensation()`
         to mitigate high-$k$ biases. See :func:`~kszx.apply_kernel_compensation()` docstring for more info.
    """

    if not isinstance(box, Box):
        raise RuntimeError("kszx.interpolate_points(): expected 'box' arg to be kszx.Box object, got {box = }")
    if kernel is None:
        raise RuntimeError("kszx.interpolate_points(): 'kernel' arg must be specified")
    if box.ndim != 3:
        raise RuntimeError('kszx.interpolate_points(): currently only ndim==3 is supported')

    arr = utils.asarray(arr, 'kszx.interpolate_points()', 'arr')
    points = utils.asarray(points, 'kszx.interpolate_points()', 'points', dtype=float)
    kernel = kernel.lower()
    
    if (points.ndim != 2) or (points.shape[1] != box.ndim):
        raise RuntimeError(f"kszx.interpolate_points(): expected points.shape=(N,{box.ndim}), got shape {points.shape}")

    is_real_space = box.is_real_space_map(arr)
    is_fourier_space = box.is_fourier_space_map(arr)

    if (not is_real_space) and (not is_fourier_space):
        raise RuntimeError("kszx.interpolate_points(): 'arr' argument has wrong shape/dtype")
    if fft and is_real_space:
        raise RuntimeError("kszx.interpolate_points(): 'arr' argument is real-space map, but fft=True was specified")
    if (not fft) and is_fourier_space:
        raise RuntimeError("kszx.interpolate_points(): 'arr' argument is Fourier-space map, you probably want to specify fft=True")

    if fft:
        arr = fft_c2r(box, arr, spin=spin)
        
    if kernel == 'cic':
        return cpp_kernels.cic_interpolate_3d(arr, points, box.lpos[0], box.lpos[1], box.lpos[2], box.pixsize, periodic)
    elif kernel == 'cubic':
        return cpp_kernels.cubic_interpolate_3d(arr, points, box.lpos[0], box.lpos[1], box.lpos[2], box.pixsize, periodic)
    else:
        raise RuntimeError(f'kszx.interpolate_points(): {kernel=} is not supported')


def _check_weights(box, points, weights, prefix=''):
    """Helper for grid_points(), used to parse (points,weights) and (rpoints,rweights) args."""

    if (points.ndim != 2) or (points.shape[1] != box.ndim):
        raise RuntimeError(f"kszx.grid_points(): expected {prefix}points.shape=(N,{box.ndim}), got shape {points.shape}")
    
    npoints = points.shape[0]
    
    if weights is None:
        weights = np.array(1.0)
    elif (weights.ndim != 0) and (weights.shape != (npoints,)):
        raise RuntimeError(f"kszx.grid_points(): {prefix}weights array has shape {weights.shape}; expected either 0-d or shape ({npoints},)")
        
    return weights


def grid_points(box, points, weights=None, rpoints=None, rweights=None, kernel=None, fft=False, spin=0, periodic=False, compensate=False, wscal=1.0):
    r"""Returns a map representing a sum of delta functions (or a "galaxies - randoms" difference map).

    Function args:

        - ``box`` (kszx.Box): defines pixel size, bounding box size, and location of observer.
          See :class:`~kszx.Box` for more info.

        - ``points`` (2-d array):
          Sequence of points where the delta functions are located (usually galaxy locations).
          Array shape should be (n,d), where n is the number of interpolation points, and d
          is the box dimension (usually 3).

        - ``weights`` (either scalar, None, or 1-d array).
           - if ``weights`` is an array, then it should have length n, where n is the number of
             interpolation points.
           - if ``weights`` is a scalar, then all delta functions have equal weight.
           - if ``weights`` is None, then all delta functions have weight 1.

        - ``rpoints`` (2-d array or None): 
          Optional sequence of points where delta functions with negative coefficients
          are located (representing a "random" catalog).

        - ``rweights`` (either scalar, None, or 1-d array).
          Weights for the ``rpoints`` array (same semantics as ``weights``).

          NOTE: an additional multiplicative normalization is applied to the ``rweights``
          so that ``sum(rweights) = -sum(weights)``. That is, if ``rweights`` are specified,
          then the (galaxies - randoms) maps returned by this function always has mean zero.

        - ``kernel`` (string): either ``'cic'`` or ``'cubic'`` (more options will be defined later).

        - ``fft`` (boolean): if True, then ``fft_r2c(.., spin)`` will be applied to the
          output map after gridding.
    
        - ``spin`` (integer): passed as 'spin' argument to ``fft_r2c()``. (Only used if ``fft=True``.)
    
        - ``compensate`` (boolean): If true, then ``kszx.apply_kernel_compensation()`` is
          called after ``fft_r2c()``. (Only used if ``fft=True``.)

        - ``periodic`` (boolean): if True, then the box has periodic boundary conditions.

        - ``wscal`` (float, default 1.0): overall scaling applied to weights.
    
    Return value: 

      - A numpy array representing a real-space (``fft=False``) or Fourier-space (``fft=True``) map.

        The real-space and Fourier-space array shapes are given by ``box.real_space_shape``
        and ``box.fourier_space_shape``, and are related as follows:

        $$\begin{align}
        (\mbox{real-space shape}) &= (n_0, n_1, \cdots, n_{d-1}) \\
        (\mbox{Fourier-space shape}) &= (n_0, n_1, \cdots, \lfloor n_{d-1}/2 \rfloor + 1)
        \end{align}$$

    Notes:

       - The ``points`` array should be specified in "observer coordinates", not "grid coordinates".

         (Reminder: in observer coordinates, the observer is at the origin, coordinates have units
         of distance, and the corners of the box are at ``box.{lpos,rpos}``. 
         See :class:`~kszx.Box` for more info.)

       - The normalization of the output map includes a factor (pixel volume)$^{-1}$. 
         (That is, the sum of the output array is $(\sum_j w_j)/V_{\rm pix}$, not $(\sum_j w_j)$.)

         This normalization best represents a weighted sum of delta functions 
         $f(x) = \sum_j w_j \delta^3(x-x_j)$. For example, if we FFT the output array with 
         ``kszx.fft_r2c()``, the result is a weighted sum of plane waves:

         $$f(k) = \sum_j w_j \exp(-i{\bf k}\cdot {\bf x_j})$$
    
         with no factor of box or pixel volume.

       - After calling ``grid_points()``, you may want to call :func:`~kszx.apply_kernel_compensation()`
         to mitigate high-$k$ biases. See :func:`~kszx.apply_kernel_compensation()` docstring for more info.
    """

    if not isinstance(box, Box):
        raise RuntimeError("kszx.grid_points(): expected 'box' arg to be kszx.Box object, got {box = }")
    if kernel is None:
        raise RuntimeError("kszx.grid_points(): 'kernel' arg must be specified")
    if (rpoints is None) and (rweights is not None):
        raise RuntimeError("kszx.grid_points(): 'rpoints' arg is None, but 'rweights' arg is not None")
    if box.ndim != 3:
        raise RuntimeError('kszx.grid_points(): currently only ndim==3 is supported')
    if (spin != 0) and (not fft):
        raise RuntimeError("kszx.grid_points(): 'spin' argument was specified with fft=False")
    if compensate and (not fft):
        raise RuntimeError("kszx.grid_points(): 'compensate' argument was specified with fft=False")

    wscal = float(wscal)
    points = utils.asarray(points, 'kszx.grid_points()', 'points', dtype=float)
    weights = utils.asarray(weights, 'kszx.grid_points()', 'weights', dtype=float, allow_none=True)
    rpoints = utils.asarray(rpoints, 'kszx.grid_points()', 'rpoints', dtype=float, allow_none=True)
    rweights = utils.asarray(rweights, 'kszx.grid_points()', 'rweights', dtype=float, allow_none=True)
    kernel = kernel.lower()
    
    if kernel == 'cic':
        cpp_kernel = cpp_kernels.cic_grid_3d
    elif kernel == 'cubic':
        cpp_kernel = cpp_kernels.cubic_grid_3d
    else:
        raise RuntimeError(f'kszx.grid_points(): {kernel=} is not supported')        
        
    grid = np.zeros(box.real_space_shape, dtype=float)
    weights = _check_weights(box, points, weights)  # also checks 'points' arg
    cpp_kernel(grid, points, weights, wscal, box.lpos[0], box.lpos[1], box.lpos[2], box.pixsize, periodic)

    if rpoints is not None:
        rweights = _check_weights(box, rpoints, rweights, prefix='r')  # also checks 'rpoints' arg
        weight_sum = len(points) * np.mean(weights) * wscal            # works if 'weights' is 0-d or 1-d
        rweight_sum = len(rpoints) * np.mean(rweights)                 # works if 'rweights' is 0-d or 1-d
        
        assert rweight_sum > 0
        rwscal = -weight_sum / rweight_sum
        cpp_kernel(grid, rpoints, rweights, rwscal, box.lpos[0], box.lpos[1], box.lpos[2], box.pixsize, periodic)

    if fft:
        grid = fft_r2c(box, grid, spin=spin)
    if compensate:
        apply_kernel_compensation(box, grid, kernel)
        
    return grid


def apply_kernel_compensation(box, arr, kernel, exponent=-0.5):
    r"""Modifies Fourier-space map 'arr' in-place, to debias interpolation/gridding.

    Context: gridding kernels (see :func:`~kszx.grid_points()`) multiplicatively bias 
    power spectrum estimation,

    $$<P(k)>_{\rm estimated} = C(k) \, P(k)_{true}$$

    Here, $C(k)$ is a "compensation factor" satisfying $0 \le C(k) \le 1$ which depends 
    on both the magnitude and orientation of $k$.

    There is a similar bias which pertains to interpolation kernels, rather than gridding
    kernels (see :func:`~kszx.interpolate_points()`). Suppose we start with a Fourier-space
    map $f(k)$, then Fourier transform and interpolate at random locations. One would
    expect that an interpolated value $f_{\rm interp}$ has variance

    $$\langle f_{\rm interp}^2 \rangle = \int \frac{d^3k}{(2\pi)^3} \, f(k)^2$$

    However, the interpolation kernel produces a bias: the actual variance is

    $$\langle f_{\rm interp}^2 \rangle = \int \frac{d^3k}{(2\pi)^3} \, C(k) f(k)^2$$

    The function ``apply_kernel_compensation`` multiplies Fourier-space map ``arr``
    in-place by $C(k)^p$, where $p$ is the ``exponent`` argument.  Here are two 
    common applications:

      1. Before calling :func:`~kszx.estimate_power_spectrum()` on one or more Fourier-space 
         maps, you should call ``apply_kernel_compensation()`` on each map, to multiply by
         $C(k)^{-1/2}$. This will mitigate the power spectrum estimation bias noted above.

      2. Before calling :func:`~kszx.interpolate_points()` on a map, you should call
         ``apply_kernel_compensation()`` on the map, to multiply by $C(k)^{-1/2}$. 
         This will mitigate the interpolation bias noted above. (This assumes that 
         you start with the map in Fourier space, and FFT before interpolating.)
    
    Function args:

        - ``box`` (kszx.Box): defines pixel size, bounding box size, and location of observer.
          See :class:`~kszx.Box` for more info.

        - ``arr``: numpy array representing a Fourier-space map. The array shape should be given by
          ``box.fourier_space_shape`` and the dtype should be ``complex``, see note below.

        - ``kernel`` (string): either ``'cic'`` or ``'cubic'`` (more options will be defined later).

        - ``exponent`` (float): array will be multiplied by ``C(k)**exponent``. (The default value
          is ``exponent = -0.5``, since this value arises in both applications above.)
    
    Return value: None (the ``arr`` argument is modified in-place, by multiplying by ``C(k)**exponent``).

    Reminder: real-space and Fourier-space array shapes are given by ``box.real_space_shape``
    and ``box.fourier_space_shape``, and are related as follows:

    $$\begin{align}
    (\mbox{real-space shape}) &= (n_0, n_1, \cdots, n_{d-1}) \\
    (\mbox{Fourier-space shape}) &= (n_0, n_1, \cdots, \lfloor n_{d-1}/2 \rfloor + 1)
    \end{align}$$
    """

    # See tex notes. The variable 's' is sin(k*L/2)
    if kernel == 'cic':
        f = lambda s: 1 - (2./3.)*s*s
    elif kernel == 'cubic':
        f = lambda s: 1 - (22./45.)*(s**4) - (124./945.)*(s**6)
    else:
        raise RuntimeError(f'kszx.gridding_pk_multiplier(): {kernel=} is not supported')

    assert isinstance(box, Box)
    assert box.is_fourier_space_map(arr)  # check shape and type of input array

    # FIXME could be optimized
    for d in range(box.ndim):
        nr = box.real_space_shape[d]
        nf = box.fourier_space_shape[d]
        s = np.sin(np.pi * np.arange(nf,dtype=float)/nr)
        arr *= np.reshape(f(s)**exponent, (1,)*d + (nf,) + (1,)*(box.ndim-d-1))


####################################################################################################


def _multiply(src, x, dest, in_place):
    """Helper for functions which take 'dest' and 'in_place' arguments.
        (multiply_rfunc(), multiply_kfunc(), multiply_r_component(), apply_partial_derivative()."""
    
    if (dest is not None) and in_place:
        raise RuntimeError("Specifying both 'dest' and 'in_place' arguments is not allowed")
    if (dest is not None) and (dest.shape != src.shape):
        raise RuntimeError("'dest' array has wrong shape")
    if (dest is not None) and (dest.shape != src.dtype):
        raise RuntimeError("'dest' array has wrong dtype")

    if (dest is src) or in_place:
        src *= x
        return src
    elif (dest is not None):
        dest[:] = src[:]
        dest *= x
        return dest
    else:
        return src * x


def _eval_kfunc(box, f, dc=None):
    """Helper for multiply_kfunc(), kbin_average()."""

    assert callable(f)
    assert isinstance(box, Box)

    k = box.get_k(regulate = (dc is not None))
    fk = f(k)

    if fk.shape != box.fourier_space_shape:
        raise RuntimeError('kszx.multiply_kfunc(): function f(k) returned unexpected shape')
    if fk.dtype != float:
        raise RuntimeError('kszx.multiply_kfunc(): function f(k) returned dtype={fk.dtype} (expected float)')

    if dc is not None:
        fk[(0,)*box.ndim] = dc

    return fk


def multiply_rfunc(box, arr, f, dest=None, in_place=False, regulate=False, eps=1.0e-6):
    r"""Multiply real-space map 'arr' by a function f(r), where r is scalar radial coordinate.

    Function args:

        - ``box`` (kszx.Box): defines pixel size, bounding box size, and location of observer.
          See :class:`~kszx.Box` for more info.
    
        - ``arr``: numpy array representing a real-space map. (The array shape should be given by
          ``box.real_space_shape`` and the dtype should be ``float``.)

        - ``f`` (function): The function r -> f(r).

        - ``dest`` (array or None): real-space map where output will be written (if None, then new array will be allocated).

        - ``in_place`` (boolean): Setting this to True is equivalent to ``dest=arr``.

        - ``regulate`` (boolean): This optional argument is intended to regulate cases
          where $\lim_{r\rightarrow 0} f(r) = \infty$. If ``regulate=True``, then we replace r
          by ``max(r, eps*pixsize)`` before calling ``f()``.

        - ``eps`` (float): only used if ``regulate=True``. See description in previous bullet point.

    Return value:

        - A real-space map. (Numpy array with same shape and dtype as ``arr``.)

    Note: 
    
       - The function ``f()`` must be vectorized: its argument 'r' will be a 3-dimensional arary,
         and the return value should be an array with the same shape.

       - r-values passed to ``f()`` will be in "observer" coordinates.

         (Reminder: in observer coordinates, the observer is at the origin, coordinates have units
         of distance, and the corners of the box are at ``box.{lpos,rpos}``. 
         See :class:`~kszx.Box` for more info.)
    """

    assert isinstance(box, Box)
    assert box.is_real_space_map(arr)   # check shape, dtype
    assert callable(f)
    assert eps < 0.5

    r = box.get_r(regulate=regulate, eps=eps)
    fr = f(r)
    
    if not box.is_real_space_map(fr):
        raise RuntimeError('kszx.multiply_rfunc(): function f(r) returned unexpected shape/dtype')

    return _multiply(arr, fr, dest, in_place)
    
    
def multiply_kfunc(box, arr, f, dest=None, in_place=False, dc=None):
    r"""Multiply Fourier-space map 'arr' by a real-valued function f(k), where k=|k| is scalar wavenumber.
    
    Function args:

        - ``box`` (kszx.Box): defines pixel size, bounding box size, and location of observer.
          See :class:`~kszx.Box` for more info.
    
        - ``arr``: numpy array representing a Fourier-space map. (The array shape should be given by
          ``box.fourier_space_shape`` and the dtype should be ``complex``, see note below.)

        - ``f`` (function): The function k -> f(k).

        - ``dest``: Fourier-space map where output will be written (if None, then new array will be allocated)

        - ``in_place``: Setting this to True is equivalent to ``dest=arr``.

        - ``dc`` (float): This optional argument is intended to regulate cases where
          $\lim_{k\rightarrow 0} f(k) = \infty$. If ``dc`` is specified, then ``f()``
          is not evaluated at k=0, and the value of ``dc`` is used instead of ``f(0)``.


    Return value:

        - A Fourier-space map. (Numpy array with same shape and dtype as ``arr``.)

    Notes: 
    
       - The function ``f()`` must be vectorized: its argument 'k' will be a 3-dimensional arary,
         and the return value should be a real-valued array with the same shape.

       - k-values passed to ``f()`` will be in "physical" units, i.e. the factor ``(2*pi / box.boxsize)``
         is included.

       - The ``arr`` argument and the returned array are Fourier-space maps.
    
         Reminder: real-space and Fourier-space array shapes are given by ``box.real_space_shape``
         and ``box.fourier_space_shape``, and are related as follows:

         $$\begin{align}
         (\mbox{real-space shape}) &= (n_0, n_1, \cdots, n_{d-1}) \\
         (\mbox{Fourier-space shape}) &= (n_0, n_1, \cdots, \lfloor n_{d-1}/2 \rfloor + 1)
         \end{align}$$
    """

    assert isinstance(box, Box)
    assert box.is_fourier_space_map(arr)   # check shape, dtype

    fk = _eval_kfunc(box, f, dc=dc)
    return _multiply(arr, fk, dest, in_place)


def multiply_r_component(box, arr, axis, dest=None, in_place=True):
    r"""Multiply real-space map 'arr' by $r_j$ (the j-th Cartesian coordinate).

    Function args:

        - ``box`` (kszx.Box): defines pixel size, bounding box size, and location of observer.
          See :class:`~kszx.Box` for more info.
    
        - ``arr``: numpy array representing a real-space map. (The array shape should be given by
          ``box.real_space_shape`` and the dtype should be ``float``.)

        - ``axis`` (integer): axis j (satisfying ``0 <= j < box.ndim``) along which $r_j$ is computed.

        - ``dest`` (array or None): real-space map where output will be written (if None, then new array will be allocated).

        - ``in_place`` (boolean): Setting this to True is equivalent to ``dest=arr``.

    Return value:

        - A real-space map. (Numpy array with same shape and dtype as ``arr``.)

    Note: 

       - Values of $r_j$ will be signed, and in "observer" coordinates.

         (Reminder: in observer coordinates, the observer is at the origin, coordinates have units
         of distance, and the corners of the box are at ``box.{lpos,rpos}``. 
         See :class:`~kszx.Box` for more info.)
    """
    
    assert isinstance(box, Box)
    assert box.is_real_space_map(arr)

    ri = box.get_r_component(axis)
    return _multiply(arr, ri, dest, in_place)

    
def apply_partial_derivative(box, arr, axis, dest=None, in_place=True):
    r"""Multiply Fourier-space map 'arr' by $(i k_j)$. (This is the partial derivative $\partial_j$ in Fourier space.)

    Function args:

        - ``box`` (kszx.Box): defines pixel size, bounding box size, and location of observer.
          See :class:`~kszx.Box` for more info.
        
        - ``arr``: numpy array representing a Fourier-space map. (The array shape should be given by
          ``box.fourier_space_shape`` and the dtype should be ``complex``, see note below.)

        - ``axis`` (integer): axis j (satisfying ``0 <= j < box.ndim``) along which $(i k_j)$ is computed.

        - ``dest`` (array or None): Fourier-space map where output will be written (if None, then new array will be allocated).

        - ``in_place`` (boolean): Setting this to True is equivalent to ``dest=arr``.

    Return value:

        - A Fourier-space map. (Numpy array with same shape and dtype as ``arr``.)

    Notes: 

       - k-values passed to ``f()`` will be in "physical" units, i.e. the factor ``(2*pi / box.boxsize)``
         is included.
    
       - Values of $k_j$ will be signed, and include the factor (2pi / boxsize).

       - The value of k_j will be taken to be zero at the Nyquist frequency.
         (I think this is the only sensible choice, since the sign is ambiguous.)

       - The ``arr`` argument and the returned array are Fourier-space maps.
    
         Reminder: real-space and Fourier-space array shapes are given by ``box.real_space_shape``
         and ``box.fourier_space_shape``, and are related as follows:

         $$\begin{align}
         (\mbox{real-space shape}) &= (n_0, n_1, \cdots, n_{d-1}) \\
         (\mbox{Fourier-space shape}) &= (n_0, n_1, \cdots, \lfloor n_{d-1}/2 \rfloor + 1)
         \end{align}$$
    """

    assert isinstance(box, Box)
    assert box.is_fourier_space_map(arr)

    ki = 1j * box.get_k_component(axis, zero_nyquist=True)
    return _multiply(arr, ki, dest, in_place)


####################################################################################################


def _to_float(x, errmsg):
    try:
        return float(x)
    except:
        raise RuntimeError(errmsg)


def _sqrt_pk(box, pk, regulate):
    """Helper for simulate_gaussian_field()."""

    if callable(pk):
        k = box.get_k(regulate=regulate)
        pk = pk(k)
        
        if pk.shape != box.fourier_space_shape:
            raise RuntimeError('kszx.simulate_gaussian_field(): function pk() returned unexpected shape')
        if pk.dtype != float:
            raise RuntimeError('kszx.simulate_gaussian_field(): function pk() returned dtype={pk.dtype} (expected float)')
        if np.min(pk) < 0:
            raise RuntimeError('kszx.simulate_gaussian_field(): function pk() returned negative values')

        del k
        pk **= 0.5
        return pk   # returns sqrt(P(k))

    pk = _to_float(pk, 'kszx.simulate_gaussian_field(): expected pk argument to be either callable, or a real scalar')

    if pk < 0:
        raise RuntimeError('kszx.simulate_gaussian_field(): expected scalar pk argument to be non-negative')
    
    return np.sqrt(pk)

    
def simulate_white_noise(box, *, fourier):
    r"""Simulate white noise, in either real space or Fourier space, normalized to $P(k)=1$.

    Intended as a helper for ``simulate_gaussian_field()``, but may be useful on its own.

    Function args:

        - ``box`` (kszx.Box): defines pixel size, bounding box size, and location of observer.
          See :class:`~kszx.Box` for more info.

        - ``fourier`` (boolean): determines whether output is real-space or Fourier-space.
    
    Return value: 

      - A numpy array representing a real-space (``fourier=False``) or Fourier-space (``fourier=True``) map.

        The real-space and Fourier-space array shapes are given by ``box.real_space_shape``
        and ``box.fourier_space_shape``, and are related as follows:

        $$\begin{align}
        (\mbox{real-space shape}) &= (n_0, n_1, \cdots, n_{d-1}) \\
        (\mbox{Fourier-space shape}) &= (n_0, n_1, \cdots, \lfloor n_{d-1}/2 \rfloor + 1)
        \end{align}$$

    Note: our normalization conventions for the simulated field are (in Fourier and real space):
    
    $$\langle f(k) f(k')^* \rangle = V_{\rm box} \delta_{kk'}$$
    $$\langle f(x) f(x') \rangle = V_{\rm pix}^{-1} \delta_{xx'}$$
    """

    if not fourier:
        rms = 1.0 / np.sqrt(box.pixel_volume)
        return np.random.normal(size=box.real_space_shape, scale=rms)
        
    # Simulate white noise in Fourier space.
    nd = box.ndim
    rms = np.sqrt(0.5 * box.box_volume)
    ret = np.zeros(box.fourier_space_shape, dtype=complex)        
    ret.real = np.random.normal(size=box.fourier_space_shape, scale=rms)
    ret.imag = np.random.normal(size=box.fourier_space_shape, scale=rms)

    # The rest of this function imposes the reality condition f(-k) = f(k)^*.
    
    # t = modes where k_{nd-1} is self-conjugate
    n = box.npix[nd-1]
    s1 = (slice(None),) * (nd-1)
    s2 = slice(0,1) if (n % 2) else slice(0, (n//2)+1, (n//2))
    tview = ret[s1+(s2,)] 
    tcopy = np.conj(tview)   # copy and complex conjugate

    # Apply parity operation k -> (-k) to 'tcopy'.
    for axis in range(nd-1):
        n = box.npix[axis]
        s1 = (slice(None),) * axis
        s2fwd = (slice(1,n),)
        s2rev = (slice(n-1,0,-1),)
        s3 = (slice(None),) * (nd-axis-1)
        u = np.copy(tcopy[s1+s2rev+s3])
        tcopy[s1+s2fwd+s3] = u

    # Replace f(k) by (f(k) - f(-k)^*) / sqrt(2)
    tview += tcopy
    tview *= np.sqrt(0.5)   # preserve variance
    return ret


def simulate_gaussian_field(box, pk, pk0=None):
    r"""Simulates a Gaussian field (in Fourier space) with specified power spectrum P(k).

    Function args:

        - ``box`` (kszx.Box): defines pixel size, bounding box size, and location of observer.
          See :class:`~kszx.Box` for more info.

        - ``pk`` (function or scalar): The power spectrum, represented as a function $k \rightarrow P(k)$.
          If the power spectrum is constant in $k$, then a scalar can be used instead of a function.

        - ``pk0`` (scalar or None): This optional argument is intended to regulate cases
          where $\lim_{k\rightarrow 0} P(k) = \infty$. If ``pk0`` is specified, then ``pk()`` is
          not evaluated at k=0, and the value of ``pk0`` is used instead of ``Pk(0)``.
    
    Return value: 

         - A numpy array representing a Fourier-space map. (Array shape is given by
           ``box.fourier_space_shape``, and dtype is complex, see note below.)

    Notes:

       - The normalization of the simulated field is:

         $$\langle f(k) f(k')^* \rangle = V_{\rm box} P(k) \delta_{kk'}$$
    
       - The function ``pk()`` must be vectorized: its argument 'k' will be a 3-dimensional arary,
         and the return value should be a real-valued array with the same shape.
    
       - k-values passed to ``pk()`` will be in "physical" units, i.e. the factor ``(2*pi / box.boxsize)``
         is included.

       - The returned array is a Fourier-space map.
    
         Reminder: real-space and Fourier-space array shapes are given by ``box.real_space_shape``
         and ``box.fourier_space_shape``, and are related as follows:

         $$\begin{align}
         (\mbox{real-space shape}) &= (n_0, n_1, \cdots, n_{d-1}) \\
         (\mbox{Fourier-space shape}) &= (n_0, n_1, \cdots, \lfloor n_{d-1}/2 \rfloor + 1)
         \end{align}$$
    """

    assert isinstance(box, Box)

    sqrt_pk = _sqrt_pk(box, pk, regulate = (pk0 is not None))
    ret = simulate_white_noise(box, fourier=True)

    dc = ret[(0,)*box.ndim]   # must precede multiplying by sqrt_pk
    ret *= sqrt_pk

    if pk0 is not None:
        pk0 = _to_float(pk0, 'kszx.simulate_gaussian_field(): expected pk0 argument to be a real scalar')
        if pk0 < 0:
            raise RuntimeError('kszx.simulate_gaussian_field(): expected pk0 argument to be non-negative')
        ret[(0,)*box.ndim] = np.sqrt(pk0) * dc
    
    return ret


####################################################################################################


def _check_kbin_edges(box, kbin_edges, use_dc):
    """Helper for estimate_power_spectrum() and kbin_average(). Returns new kbin_edges."""

    kbin_edges = np.asarray(kbin_edges, dtype=float)
    
    assert isinstance(box, Box)
    assert kbin_edges.ndim == 1
    assert len(kbin_edges) >= 2    
    assert kbin_edges[0] >= 0.
    assert utils.is_sorted(kbin_edges)
        
    if (not use_dc) and (kbin_edges[0] == 0):
        kbin_edges = np.copy(kbin_edges)
        kbin_edges[0] = min(np.min(box.kfund), kbin_edges[1]) / 2.

    return kbin_edges
    
    
def _parse_map_or_maps(box, map_or_maps, caller):
    """Helper for estimate_power_spectrum(). Returns (map_list, multi_map_flag)."""

    if not isinstance(box, Box):        
        raise RuntimeError(f"{caller}(): expected 'box' argument to an instance of class kszx.Box")
    
    if box.is_fourier_space_map(map_or_maps):
        return ([map_or_maps], False)  # single map

    try:
        map_list = list(map_or_maps)
        if all(box.is_fourier_space_map(x) for x in map_list):
            return (map_list, True)
    except:
        pass

    raise RuntimeError(f"{caller}(): expected argument to be either a Fourier-space map,"
                       + " or an iterable returning Fourier-space maps")
    

def estimate_power_spectrum(box, map_or_maps, kbin_edges, *, use_dc=False, allow_empty_bins=False, return_counts=False):
    r"""Computes power spectrum $P(k)$ for one or more maps (including cross-spectra). The window function is not deconvolved.

    Function args:

        - ``box`` (kszx.Box): defines pixel size, bounding box size, and location of observer.
          See :class:`~kszx.Box` for more info.

        - ``map_or_maps`` (array or list of arrays): single or multiple Fourier-space maps.

           - If ``map_or_maps`` is an array, then it represents a single Fourier-space map.
             (The array shape should be given by ``box.fourier_space_shape`` and the dtype should be 
             ``complex``, see note below.)
    
           - If ``map_or_maps`` is a list of arrays, then it represents multiple Fourier-space maps.
             (Each map in the list should have shape ``box.fourier_space_shape`` and dtype ``complex``,
             see note below.)

        - ``kbin_edges`` (1-d array): 1-d array of length (nkbins+1) defining bin endpoints.
          The i-th bin covers k-range ``kbin_edges[i] <= i < kbin_edges[i+1]``.

        - ``use_dc`` (boolean): if False (the default), then the k=0 mode will not be used,
          even if the lowest bin includes k=0.

        - ``allow_empty_bins`` (boolean): if False (the default), then an execption is thrown 
          if a k-bin is empty.

        - ``return_counts`` (boolean): See below.

    Return value: 

       - An array ``pk`` is returned, with two cases as follows:

           - If the ``map_or_maps`` argument is a single Fourier-space map (see above),
             then ``pk`` is a 1-d array with length ``nkbins``, containing binned power
             spectrum estimates.

           - If the ``map_or_maps`` argument is a list of Fourier-space maps (see above),
             then ``pk`` is a 3-d array with shape ``(nmaps, nmaps, nkbins)``, containing
             all auto- and cross-power spectrum estimates.

       - If ``return_counts=False`` (the default), then the return value is simply the ``pk`` array.

         If ``return_counts=True``, then the return value is a pair ``(pk, bin_counts)``, where
         ``bin_counts`` is a 1-d array with length ``nkbins``, containing the number of Fourier
         modes in each k-bin.

    Notes:

       - Normalization: to estimate the power spectrum, we square each Fourier mode and divide by 
         the box volume. This is consistent with our normalization for Fourier-space maps (see e.g.
         the :func:`~kszx.simulate_gaussian_field` docstring), which is:

         $$\langle f(k) f(k')^* \rangle = V_{\rm box} P(k) \delta_{kk'}$$
    
       - The normalization of the estimated power spectrum P(k) assumes that the maps fill the
         entire box volume. If this is not the case (i.e. there is a survey window) then you'll
         need to renormalize P(k) or deconvolve the window function.

       - The input arrays are Fourier-space maps.
    
         Reminder: real-space and Fourier-space array shapes are given by ``box.real_space_shape``
         and ``box.fourier_space_shape``, and are related as follows:

         $$\begin{align}
         (\mbox{real-space shape}) &= (n_0, n_1, \cdots, n_{d-1}) \\
         (\mbox{Fourier-space shape}) &= (n_0, n_1, \cdots, \lfloor n_{d-1}/2 \rfloor + 1)
         \end{align}$$

       - Before calling ``estimate_power_spectrum()``, you may want to call :func:`~kszx.apply_kernel_compensation()`
         to mitigate high-$k$ biases. See :func:`~kszx.apply_kernel_compensation()` docstring for more info.
    """

    kbin_edges = _check_kbin_edges(box, kbin_edges, use_dc)
    map_list, multi_map_flag = _parse_map_or_maps(box, map_or_maps, 'kszx.estimate_power_spectrum')
    pk, bin_counts = cpp_kernels.estimate_power_spectrum(map_list, kbin_edges, box.npix, box.kfund, box.box_volume)

    if (not allow_empty_bins) and (np.min(bin_counts) == 0):
        raise RuntimeError('kszx.estimate_power_spectrum(): some k-bins were empty')
    
    if not multi_map_flag:
        pk = pk[0,0,:]   # shape (1,1,nkbins) -> shape (nkbins,)

    return (pk, bin_counts) if return_counts else pk


def kbin_average(box, f, kbin_edges, *, use_dc=False, allow_empty_bins=False, return_counts=False):
    """Averages a real-valued array/function f(k) in k-bins.

    Function args:

        - ``box`` (kszx.Box): defines pixel size, bounding box size, and location of observer.
          See :class:`~kszx.Box` for more info.

        - ``f`` (function or array): the quantity f(k) to be averaged, represented either
          as a real-valued array of shape ``box.fourier_space_shape``, or a function k -> f(k).

        - ``kbin_edges``: 1-d array of length (nkbins+1) defining bin endpoints.
          The i-th bin covers k-range kbin_edges[i] <= k < kbin_edges[i+1].

        - ``use_dc`` (boolean): if False (the default), then the k=0 mode will not be used,
          even if the lowest bin includes k=0.

        - ``allow_empty_bins`` (boolean): if False (the default), then an execption is thrown 
          if a k-bin is empty.

        - ``return_counts`` (boolean): See below.

    Return value: 

       - If ``return_counts=False`` (the default), then return value is a 1-d array ``fk_mean``
         with length nkbins, containing bin-averaged values of f(k).

         If ``return_counts=True``, then the return value is a pair ``(fk_mean, bin_counts)``, where
         ``bin_counts`` is a 1-d array with length ``nkbins``, containing the number of Fourier
         modes in each k-bin.

    Notes: 

       - This function is intended to be used in situations where we want to compare the
         output of estimate_power_spectrum() to a "theory" power spectrum, such as ``Cosmology.plin_z0()``.
         To remove binning artifacts, you may want to bin-average the theory power spectrum
         over the same k-bins used in estimate_power_spectrum().
    
       - The function f() must be vectorized: its argument 'k' will be a 3-dimensional arary,
         and the return value should be a real-valued array with the same shape.

       - k-values passed to f() will include the factor (2pi / boxsize).

       - The output of kbin_average() does not include any normalization factors (such as the
         box or pixel size). It is just a straightforward average of f() values over each k-bin.
    """

    kbin_edges = _check_kbin_edges(box, kbin_edges, use_dc)

    if callable(f):
        fk = _eval_kfunc(box, f, dc = (None if use_dc else 0.))
        assert fk.shape == box.fourier_space_shape
        assert fk.dtype == float
    else:
        fk = utils.asarray(f, 'kszx.kbin_edges', 'f', dtype=float)
        assert fk.shape == box.fourier_space_shape

    fk_mean, bin_counts = cpp_kernels.kbin_average(fk, kbin_edges, box.npix, box.kfund)

    if (not allow_empty_bins) and (np.min(bin_counts) == 0):
        raise RuntimeError('kszx.kbin_average(): some k-bins were empty')

    return (fk_mean, bin_counts) if return_counts else fk_mean


####################################################################################################


def fkp_from_ivar_2d(ivar, cl0, normalize=True, return_wvar=False):
    """Makes a 2-d pixell FKP map (intended for CMB) from a pixell inverse variance map.

    The 'ivar' argument is an inverse noise varaiance map (i.e. pixell.enmap) in (uK)^{-2},
    for example the return value of act.read_ivar().

    The 'cl0' argument parameterizes the FKP weight function.
      - cl0=0 corresponds to inverse noise weighting.
      - Large cl0 corresponds to uniform weighing (but cl0=inf won't work).
      - Intuitively, cl0 = "fiducial signal C_l at wavenumber l of interest".
      - I usually use cl0 = 0.01 for plotting all-sky CMB temperature maps.
      - I usually use cl0 ~ 10^(-5) for small-scale KSZ filtering.

    The FKP weight function is defined by

      W(x) = 1 / (Cl0 + Nl(x))     if normalize=False

    where Nl(x) = pixarea(x) / ivar(x) is the "local" noise power spectrum. In implementation, 
    in order to avoid divide-by-zero for Cl0=0 or ivar=0, we compute W(x) equivalently as:

      W(x) = ivar(x) / (pixarea(x) + Cl0 * ivar(x))         if normalize=False

    If normalize=True, then we normalize the weight function so that max(W)=1.
    
    If wvar=True, then we return W(x) var(x) = W(x) / ivar(x), instead of returning W(x).
    """
    
    assert isinstance(ivar, pixell.enmap.ndmap)
    assert ivar.ndim == 2
    assert np.all(ivar >= 0.0)
    assert cl0 >= 0.0

    wvar = 1.0 / (ivar.pixsizemap() + cl0 * ivar)
    w = wvar * ivar
    
    wmax = np.max(w)
    assert wmax > 0.0

    ret = wvar if return_wvar else w
    
    if normalize:
        ret /= wmax

    return ret


def ivar_combine(ivar1, ivar2):
    """Given two 2-d pixell ivar maps, return the 'combined' ivar map (1/ivar1 + 1/ivar2)^{-1}."""
    
    assert isinstance(ivar1, pixell.enmap.ndmap)
    assert isinstance(ivar2, pixell.enmap.ndmap)

    den = ivar1 + ivar2    
    den = np.where(den > 0, den, 1.0)
    num = ivar1 * ivar2
    
    num /= den
    return num


def estimate_cl(alm_or_alms, lbin_delim):
    """Similar interface to estimate_power_spectrum(), but for all-sky alms.

    It's okay if lbin_delim is a float array (it will be converted to int).

    If 'alm_or_alms' is a 1-d array (single alm), returns a 1-d array of length (nlbins,).
    If 'alm_or_alms' is a 2-d array (multiple alms), returns a 3-d array of shape (nalm,nalm,nlbins).
    """

    alm_or_alms = utils.asarray(alm_or_alms, 'kszx.estimate_cl()', 'alm_or_alms')
    lbin_delim = utils.asarray(lbin_delim, 'kszx.estimate_cl()', 'lbin_delim', dtype=int)
    multi_flag = True

    # Check 'lbin_delim' arg.
    if lbin_delim.ndim != 1:
        raise RuntimeError("kszx.estimate_cl(): expected 'lbin_delim' to be a 1-d array")
    if len(lbin_delim) < 2:
        raise RuntimeError("kszx.estimate_cl(): expected 'lbin_delim' to have length >= 2")
    if not utils.is_sorted(lbin_delim):
        raise RuntimeError("kszx.estimate_cl(): 'lbin_delim' was not sorted, or bins were too narrow")
    if lbin_delim[0] < 0:
        raise RuntimeError("kszx.estimate_cl(): expected 'lbin_delim' elements to be >= 0")
    
    # Check 'alm_or_alms' arg and convert to 2-d.
    if (alm_or_alms.dtype != complex) and (alm_or_alms.dtype != np.complex64):
        raise RuntimeError("kszx.estimate_cl(): 'alm_or_alms' array did not have complex dtype")
    if alm_or_alms.ndim == 1:
        alm_or_alms = np.reshape(alm_or_alms, (1,-1))
        multi_flag = False
    elif alm_or_alms.ndim != 2:
        raise RuntimeError("kszx.estimate_cl(): 'alm_or_alms' array did not have expected shape")

    alm_or_alms = np.asarray(alm_or_alms, dtype=complex)   # convert complex64 -> complex128
    nalm, nlm = alm_or_alms.shape
    lmax = int(np.sqrt(2*nlm) - 1)

    if (nlm == 0) or ((2*nlm) != (lmax+1)*(lmax+2)):
        raise RuntimeError("kszx.estimate_cl(): 'alm_or_alms' array did not have expected shape")
    if (lbin_delim[-1] > lmax):
        raise RuntimeError(f"kszx.estimate_cl(): l-bin endpoint (={lbin_delim[-1]}) was > alm_lmax (={lmax})")
    
    nlbins = len(lbin_delim) - 1
    cl = np.zeros((nalm,nalm,lmax+1))
    ret = np.zeros((nalm,nalm,nlbins))
    
    # FIXME I think this can be done with one call to pixell.alm2cl().
    for i in range(nalm):
        for j in range(i+1):
            cl[i,j,:] = cl[j,i,:] = healpy.alm2cl(alm_or_alms[i,:], alm_or_alms[j,:])

    for b in range(nlbins):
        bin_lmin, bin_lmax = lbin_delim[b], lbin_delim[b+1]
        ret[:,:,b] = np.mean(cl[:,:,bin_lmin:bin_lmax], axis=2)

    if not multi_flag:
        ret = ret[0,0,:]

    return ret



import copy
import camb
import numpy as np
import scipy.interpolate
import matplotlib.pyplot as plt

from . import utils   # is_sorted(), spline1d(), spline2d()


####################################################################################################


class CosmologicalParams:
    def __init__(self, name=None):
        r"""Simple data class containing cosmological params (ombh2, omch2, etc.)

        You probably don't need to use this class -- instead construct a :class:`~kszx.Cosmology`
        object by name.

        Constructor args:
        
           - name (string or None).

             The following names are supported:
                 - ``params='planck18+bao'``: https://arxiv.org/abs/1807.06209 (last column of Table 2)
                 - ``params='hmvec'``: match defaults in Mat's hmvec code (https://github.com/simonsobs/hmvec)

        If no name is specified, then caller must set ombh2, omch2, etc. after calling constructor.
        """
        
        self.ombh2 = None
        self.omch2 = None
        self.scalar_amp = None
        self.tau = None
        self.ns = None
        self.h = None

        self.kpivot = 0.05
        self.nnu = 3.046
        self.mnu = 0.06
        self.zmax = 10.0
        self.kmax = 100.0
        self.dz = 0.1
        self.lmax = 5000   # reasonable default?
        self.Yp = 0.24

        # Accuracy settings
        # See sec 5 of https://arxiv.org/abs/2103.05582 for comments on high-l CMB
        
        self.accuracy_boost = 1.0
        self.l_sample_boost = 1.0
        self.l_accuracy_boost = 1.0
        self.do_late_rad_truncation = True
        self.lens_potential_accuracy = 8   # This parameter doesn't affect running time much
        self.lens_margin = 2050            # This parameter doesn't affect running time much
        
        if name == 'planck18+bao':
            # Reference: https://arxiv.org/abs/1807.06209 (last column of Table 2)
            self.ombh2 = 0.02242  # \pm 0.00014
            self.omch2 = 0.11933  # \pm 0.00091
            self.scalar_amp = 2.105e-9  # \pm 0.03
            self.tau = 0.0561     # \pm 0.0071
            self.ns = 0.9665      # \pm 0.0038
            self.h = 0.6766       # \pm 0.42
        elif name == 'hmvec':
            # Match defaults in Mat's hmvec code.
            # Reference: https://github.com/simonsobs/hmvec/blob/master/hmvec/params.py
            self.ombh2 = 0.02225
            self.omch2 = 0.1198
            self.scalar_amp = 2.2e-9
            self.tau = 0.06
            self.ns = 0.9645
            self.h = 0.673
            self.mnu = 0.0   # note zero neutrino mass
        elif name is not None:
            raise RuntimeError(f"CosmologicalParams: name '{name}' not recognized")

        
    def validate(self):
        for k in [ 'ombh2', 'omch2', 'scalar_amp', 'tau', 'ns', 'h' ]:
            if getattr(self, k) is None:
                raise RuntimeError(f"Must either initialize CosmologicalParams.{k}, or call e.g. CosmologicalParams('planck18+bao')")

        # Catch accidental use of 100*h instead of h.
        assert self.h < 2.0

        # FIXME could add more checks here, e.g.
        #  assert self.ombh2 > 0
        #  assert self.zmax > 0


####################################################################################################

        
class Cosmology:
    def __init__(self, params, lmax=None):
        r"""Thin wrapper around CAMB 'results' object, with methods such as H(), Plin(), etc.

        Adding a wrapper around CAMB is not really necessary, but I like it for a few reasons:

           - The Cosmology object is pickleable (unlike the camb 'results' object).
           - I find the syntax a little more intuitive.
           - It's a convenient place to add new methods (e.g. frsd(), alpha()).

        NOTE: no h-units! All distances are Mpc (not $h^{-1}$ Mpc), and all wavenumbers 
        are Mpc$^{-1}$ (not h Mpc$^{-1}$). All masses are Msol.

        Constructor args:
        
           - params (:class:`~kszx.CosmologicalParams` object, or string name).

             The following string names are supported:
                 - ``params='planck18+bao'``: https://arxiv.org/abs/1807.06209 (last column of Table 2)
                 - ``params='hmvec'``: match defaults in Mat's hmvec code (https://github.com/simonsobs/hmvec)

           - lmax (int, optional): if specified, overrides the value of lmax in 'params'.

        Note: methods require caller to specify keywords, e.g. caller must call ``Cosmology.Plin(k=xx, z=xx)``
        instead of ``Plin(xx,xx)``. This is intentional, to reduce the chances that I'll create a bug by swapping
        arguments or using the wrong time coordinate (e.g. scale factor `a` instead of redshift `z`).
        """
        
        if isinstance(params, str):
            params = CosmologicalParams(params)   # 'params' arg is a string, e.g. 'planck18+bao' or 'hmvec'.
        elif not isinstance(params, CosmologicalParams):
            raise RuntimeError("Cosmology constructor: argument must be either a CosmologicalParams, or a string e.g. 'planck18+bao'")

        if lmax is not None:
            params = copy.copy(params)
            params.lmax = lmax
            
        params.validate()
        
        self.params = params
        self.zmax = params.zmax
        self.kmax = params.kmax
        self.lmax = params.lmax
        self.h = params.h

        # Some unit conversions:
        #
        # u = pint.UnitRegistry()
        # u.define('solar_mass = 1.989e+30 * kg = Msol')
        # G = u.gravitational_constant
        # rho_crit_over_h2 = 3/(8*np.pi*G) * (100*u("km/sec/Mpc"))**2
        # print(rho_crit_over_h2.to("Msol/Mpc^3"))
        # print((1*u.proton_mass).to("Msol"))

        rho_crit_over_h2 = 2.7754e11   # Msol / Mpc^3
        mp = 8.40936e-58               # proton mass in Msol

        # self.rhom_comoving = comoving matter density (z-independent), units Msol Mpc^{-3}
        self.rhom_comoving = rho_crit_over_h2 * (params.ombh2 + params.omch2)   # (Msol/Mpc^3)

        # self.ne0 = electron density at z=0, units Mpc^{-3}.
        self.ne0 = (1-self.params.Yp/2) * rho_crit_over_h2 * params.ombh2 / mp

        # CAMB code below loosely follows:
        # https://camb.readthedocs.io/en/latest/CAMBdemo.html
        
        camb_params = camb.CAMBparams()
        
        camb_params.set_cosmology(
            H0 = 100 * params.h,
            ombh2 = params.ombh2,
            omch2 = params.omch2,
            nnu = params.nnu,
            mnu = params.mnu,
            tau = params.tau
        )
        
        camb_params.InitPower.set_params(
            As = params.scalar_amp,
            ns = params.ns
        )

        camb_params.set_for_lmax(
            params.lmax,
            lens_potential_accuracy = params.lens_potential_accuracy,
            lens_margin = params.lens_margin
        )

        camb_params.set_accuracy(
            AccuracyBoost = params.accuracy_boost,
            lSampleBoost = params.l_sample_boost,
            lAccuracyBoost = params.l_accuracy_boost,
            DoLateRadTruncation = params.do_late_rad_truncation
        )

        # Options here are: (NonLinear_none, NonLinear_pk, NonLinear_lens, NonLinear_both)
        camb_params.NonLinear = camb.model.NonLinear_both
        
        nz_pk = int(round(params.zmax / params.dz)) + 1
        nz_pk = max(nz_pk, 2)
        z_pk = np.linspace(params.zmax, 0, nz_pk)
        
        camb_params.set_matter_power(redshifts=z_pk, kmax=params.kmax)

        print(f'Running CAMB')
        camb_results = camb.get_results(camb_params)

        # self._sigma8_z = 1-d array of redshifts (ordered from highest to lowest)
        # self._sigma8 = sigma8 value at each redshift
        self._sigma8_z = z_pk
        self._sigma8 = camb_results.get_sigma8()

        # CMB power spectra
        
        powers = camb_results.get_cmb_power_spectra(camb_params, CMB_unit='muK', raw_cl=True)
        
        lmax = self.lmax
        self.cltt_unl = powers['unlensed_total'][:(lmax+1),0].copy()
        self.cltt_len = powers['total'][:(lmax+1),0].copy()
        self.clphi = powers['lens_potential'][:(lmax+1),0].copy()

        # Large-scale structure power spectra
        
        k, z, pzk = camb_results.get_linear_matter_power_spectrum(var1=None, var2=None, hubble_units=False, k_hunit=False, nonlinear=False)
        self._plin_k = k
        self._plin_z = z
        self._plin_pzk = pzk

        assert utils.is_sorted(k)
        assert utils.is_sorted(z)
        assert k[0] > 0.0
        assert k[-1] >= self.params.kmax
        assert z[0] == 0.0
        assert np.all(pzk > 0.0)
        assert pzk.shape == (len(z), len(k))

        # Q(k,z) = log(P(k,z)/k). Note that we transpose the (k,z) axes relative to pzk.
        Q = np.log(pzk.T / k.reshape((-1,1)))
        
        self._pk_kmin = k[0]
        self._pk0_interp = utils.spline1d(np.log(k), np.copy(Q[:,0]))  # interpolate log(k) -> Q
        self._pkz_interp = utils.spline2d(np.log(k), z, Q)             # interpolate (log(k),z) -> Q

        # Growth function.
        self._Dfit_z0 = self.Dfit(z=0, z0norm=False)
        self._Dfit_zhi = self.Dfit(z=z[-1], z0norm=False)
        ik = np.searchsorted(k, 0.1)
        # print(f'Inferring growth function D(z) from P(k,z) at k={k[ik]}')
        logD = 0.5 * np.log(pzk[:,ik])
        logD -= logD[0]
        self._logD_interp = utils.spline1d(z, logD)           # interpolate z -> log(D(z)), normalized so that D(0)=1.
        self._logD_shift = np.log(self._Dfit_zhi) - logD[-1]  # log(D) shift to normalize so that D(z) -> 1/(1+z) at high z.

        # Background expansion
        
        nz = 1000
        zvec = np.linspace(0.0, self.zmax, nz)
        Hvec = camb_results.h_of_z(zvec)
        chivec = camb_results.comoving_radial_distance(zvec)

        # Define R = chi/z
        Rvec = np.zeros(nz)
        Rvec[1:] = chivec[1:] / zvec[1:]
        Rvec[0] = 1.0 / Hvec[0]

        self.chimax = chivec[-1]
        self._Hz_interp = utils.spline1d(zvec, Hvec)           # interpolate zvec -> Hvec
        self._Rz_interp = utils.spline1d(zvec, Rvec)           # interpolate z -> (chi/z)
        self._Rchi_interp = utils.spline1d(chivec, 1.0/Rvec)   # interpolate chi -> (z/chi)
        

    def H(self, *, z, check=True):
        r"""Returns Hubble expansion rate $H(z)$."""
        
        if check:
            assert np.all(z >= 0.0)
            assert np.all(z <= self.zmax * (1.0+1.0e-8))
        
        return self._Hz_interp(z)   # Hz_interp interpolates z -> H(z)


    def chi(self, *, z, check=True):
        r"""Returns comoving distance $\chi(z)$."""

        z = np.asarray(z)
        
        if check:
            assert np.all(z >= 0.0)
            assert np.all(z <= self.zmax * (1.0+1.0e-8))
        
        return z * self._Rz_interp(z)   # Rz_interp interpolates z -> (chi/z)

    
    def z(self, *, chi, check=True):
        r"""Returns redshift $z$ corresponding to specified comoving distance $\chi$."""

        chi = np.asarray(chi)
        
        if check:
            assert np.all(chi >= 0.0)
            assert np.all(chi <= self.chimax * (1.0+1.0e-8))
        
        return chi * self._Rchi_interp(chi)   # Rchi_interp interpolates chi -> (z/chi)


    def Plin_z0(self, k, check=True):
        r"""Returns linear power spectrum $P(k)$ at $z=0$. 

        Calling ``Plin_z0()`` is slightly faster than calling ``Plin(k,z)`` with k=0,
        and in most situations, calling ``Plin(k,z)`` is unnecessary, since:

        $$P_{lin}(k,z) \approx P_{lin}(k,0) \frac{D(z)}{D(0)}$$

        This approximation slightly breaks down (at the ~0.5% level!)
        on large scales $k \lesssim 10^{-3}$ and small scales $k \gtrsim 0.1$.
        """

        k = np.asarray(k)
        
        if check:
            assert np.all(k >= 0.0)
            assert np.all(k <= self.kmax * (1.0+1.0e-8))

        kk = np.maximum(k, self._pk_kmin)
        Q = self._pk0_interp(np.log(kk))   # Q = log(P(k)/k)
        return k * np.exp(Q)


    def Plin(self, *, k, z, kzgrid=False, check=True):
        r"""Returns linear power spectrum $P(k,z)$.

        The ``kzgrid`` argument has the following meaning:

          - ``kzgrid=False`` means "broadcast k,z to get a set of points"
          - ``kzgrid=True``   means "take the Cartesian product of k,z to get a kzgrid"

        (If ``kzgrid=True``, then the returned array has shape (nk,nz) -- note that this
        convention is transposed relative to CAMB or hmvec.)
        """

        k, z = np.asarray(k), np.asarray(z)
        
        if check:
            assert np.all(k >= 0.0)
            assert np.all(k <= self.kmax * (1.0+1.0e-8))
            assert np.all(z >= 0.0)
            assert np.all(z <= self.zmax * (1.0+1.0e-8))

        # Note: the 'grid' argument to scipy.interpolate.RectBivariateSpline.__call__()
        # has the semantics as our 'kzgrid'.
        
        kk = np.maximum(k, self._pk_kmin)
        Q = self._pkz_interp(np.log(kk), z, grid=kzgrid)   # Q = log(P(k)/k)

        if kzgrid:
            k = k.reshape(k.shape + ((-1,)*z.ndim))
        
        return k * np.exp(Q)
    

    def D(self, *, z, z0norm, check=True):
        r"""Return the growth function $D(z)$.

        If ``z0norm=True``, normalize so that $D(0)=1$.

        If ``z0norm=False``, normalize so that $D(z) \rightarrow 1/(1+z)$ at high z."""

        z = np.asarray(z)
        
        if check:
            assert np.all(z >= 0.0)
            assert np.all(z <= self.zmax * (1.0+1.0e-8))

        logD = self._logD_interp(z)
        if not z0norm:
            logD += self._logD_shift
        
        return np.exp(logD)

    
    def Dfit(self, *, z, z0norm, check=True):
        r"""Return the growth function ``D(z)``, computed using a popular fitting function.

        This is probably not the function you want! You probably want ``D()``, not ``Dfit()``."""

        z = np.asarray(z)
        
        if check:
            assert np.all(z >= 0.0)
            assert np.all(z <= self.zmax * (1.0+1.0e-8))
            
        omm0 = (self.params.ombh2 + self.params.omch2) / self.h**2
        oml0 = 1.0 - omm0
        omlz = oml0 / (oml0 + omm0 * (1+z)**3)
        ommz = 1.0 - omlz
        D = 2.5 * ommz / (ommz**(4.0/7.0) - omlz + (1.0+ommz/2.0)*(1.0+omlz/70.0)) / (1.0+z)
        return (D / self._Dfit_z0) if z0norm else D

    
    def frsd(self, *, z, check=True):
        r"""Return RSD function $f(z) = d(\log D)/d(\log a)$. Uses a fitting function for now!"""

        z = np.asarray(z)
        
        if check:
            assert np.all(z >= 0.0)
            assert np.all(z <= self.zmax * (1.0+1.0e-8))
        
        # FIXME using fitting function for now
        omm0 = (self.params.ombh2 + self.params.omch2) / self.h**2
        oml0 = 1.0 - omm0
        ommz = omm0 / (omm0 + oml0 * (1+z)**(-3.))
        return ommz**(5./9.)

        
    def alpha(self, *, k, z, kzgrid=False, check=True):
        r"""Return $\alpha(k,z)$, defined by $\delta_m(k,z) = (3/5) \alpha(k,z) \zeta(k)$.

        The function $\alpha(k,z)$ arises in non-Gaussian halo bias as:

        $$b(k,z) = b_g + 2 b_{ng} \frac{f_{NL}}{\alpha(k,z)}$$

        where $b_{ng} = d(\log n)/d(\log \sigma_8) \approx \delta_c (b_g - 1)$.

        Note that for fixed k, $\alpha(k,z)$ is proportional to $D(z)$, to an excellent
        approximation.

        The ``kzgrid`` argument has the following meaning:

          - ``kzgrid=False`` means "broadcast k,z to get a set of points"
          - ``kzgrid=True`` means "take the Cartesian product of k,z to get a kzgrid"

        (If ``kzgrid=True``, then the returned array has shape (nk,nz) -- note that this
        convention is transposed relative to CAMB or hmvec.)
        """

        # Easiest way to compute alpha(k,z):
        #  alpha(k,z) = (5/3) (Pm(k,z) / Pzeta(k))**0.5
        
        k, z = np.asarray(k), np.asarray(z)
        Pm = self.Plin(k=k, z=z, kzgrid=kzgrid, check=check)
        Pz_rec = self._pzeta_rec(k, check=check)
        
        if kzgrid:
            Pz_rec = Pz_rec.reshape(Pz_rec.shape + (-1,)*z.ndim)

        return (5/3.) * np.sqrt(Pm * Pz_rec)


    def alpha_z0(self, k, check=True):
        r"""Returns $\alpha(k)$ at $z=0$, defined by $\delta_m(k) = (3/5) \alpha(k) \zeta(k)$.

        Calling ``alpha_z0()`` is slightly faster than calling ``alpha(k,z)`` with k=0,
        and in most situations, calling ``alpha(k,z)`` is unnecessary, since:

        $$\alpha(k,z) \approx \alpha(k,0) \frac{D(z)}{D(0)}$$

        The function $\alpha(k,z)$ arises in non-Gaussian halo bias as:

        $$b(k,z) = b_g + 2 b_{ng} \frac{f_{NL}}{\alpha(k,z)}$$

        where $b_{ng} = d(\log n)/d(\log \sigma_8) \approx \delta_c (b_g - 1)$.
        """
        
        # Easiest way to compute alpha(k,z):
        #  alpha(k,z) = (5/3) (Pm(k,z) / Pzeta(k))**0.5

        Pm = self.Plin_z0(k, check=True)
        Pz_rec = self._pzeta_rec(k, check=True)
        return (5/3.) * np.sqrt(Pm * Pz_rec)


    def _pzeta_rec(self, k, check=True):
        """Returns 1/P_zeta(k). Intended as a helper for alpha()."""

        if check:
            assert np.min(k) >= 0.0
            
        kpiv = self.params.kpivot
        Delta2 = self.params.scalar_amp
        ns = self.params.ns
        
        # Pzeta(k) = (2pi^2 / k^3) * Delta^2 * (k/kpiv)**(ns-1)
        # 1/Pzeta(k) = (kpiv^3 / (2 pi^2 Delta^2)) * (k/kpiv)**(4-ns)

        # Okay for k=0 (provided ns < 4!)
        return (kpiv**3 / (2 * np.pi**2 * Delta2)) * (k/kpiv)**(4-ns)
        

    def K(self, *, z):
        """Returns the kSZ radial weight K(z) = -T_CMB sigma_T n_{e0} x_e(z) e^{-tau(z)} (1+z)^2 in units (uK/Mpc).
        
        For now, we approximate x_e=1 and tau=0. Then K(z) is just proportional to (1+z)^2. I'll improve this later!
        Note that K(z) is negative.
        """

        # FIXME put tcmb in CosmologicalParams?
        tcmb = 2.726e6   # uK

        # u = pint.UnitRegistry()
        # print(float(u.thomson_cross_section / u("Mpc^2")))
        sigmaT = 6.986845e-74
        
        return -tcmb * sigmaT * self.ne0 * (1+z)**2
        



"""
The ``kszx.pixell_utils`` module contains wrappers around the pixell library.

References:
   - https://github.com/simonsobs/pixell
   - https://pixell.readthedocs.io/en/latest
   - https://github.com/simonsobs/pixell_tutorials
"""

import numpy as np
import pixell.enmap
import pixell.enplot
import pixell.curvedsky

from . import utils
from . import io_utils

from .Catalog import Catalog


def read_map(filename):
    """Reads pixell map in FITS format, and returns a pixell.enmap.
    
    If the FITS file contains temperature + polarization, only temperature will be returned!
    """
    print(f'Reading {filename}\n', end='')
    assert filename.endswith('.fits')
    return pixell.enmap.read_map(filename)


def write_map(filename, m):
    """Writes pixell map in FITS format."""
    
    # I learned the hard way that you need this assert! If pixell.enmap.write_map() is
    # called with a numpy array (instead of a pixell ndmap), it will write a "time-bomb"
    # WCS which causes some pixell functions to fail later, with cryptic error messages.
    assert isinstance(m, pixell.enmap.ndmap)
    
    print(f'Writing {filename}\n', end='')
    assert filename.endswith('.fits')
    io_utils.mkdir_containing(filename)
    pixell.enmap.write_map(filename, m)


def plot_map(m, downgrade, nolabels=True, filename=None, **kwds):
    """Plots pixell map 'm'.

    Thin wrapper around ``pixell.enplot()``, just adding a few tweaks:

       - Make downgrade argument mandatory (usually needed in practice to avoid
         runaway heavyweight behavior). Suggest ``downgrade=20`` for ACT.

       - Make ``nolabels=True`` the default (I haven't figured out how to make pixell
         labels look reasonable).

       - Add 'filename' argument to select between show/write.

    For more kwds, see:
      https://pixell.readthedocs.io/en/latest/reference.html#pixell.enplot.plot"""
    
    assert isinstance(m, pixell.enmap.ndmap)

    bunch = pixell.enplot.plot(m, downgrade=downgrade, nolabels=nolabels, **kwds)

    if filename is None:
        pixell.enplot.show(bunch)
    else:
        print(f'Writing {filename}\n', end='')
        assert filename.endswith('.png')
        io_utils.mkdir_containing(filename)
        pixell.enplot.write(filename[:-4], bunch)


####################################################################################################


def _accum_catalog(ret, shape, wcs, catalog, weights, allow_outliers, normalize_sum=None):
    """Helper for map_from_catalog(). Returns sum of weights."""

    if not allow_outliers:
        # Note: if allow_outliers is False, then pixell_utils.ang2pix() returns (idec, ira).
        msg = "kszx.pixell_utils.map_from_catalog(): some points were out-of-bounds, you may want allow_outliers=True"
        idec, ira = ang2pix(shape, wcs, catalog.ra_deg, catalog.dec_deg, allow_outliers=False, outlier_errmsg=msg)
    else:
        # Note: if allow_outliers is True, then pixell_utils.ang2pix() returns (idec, ira, mask).
        idec, ira, mask = ang2pix(shape, wcs, catalog.ra_deg, catalog.dec_deg, allow_outliers=True)
        idec, ira = idec[mask], ira[mask]
        wvec_flag = (weights is not None) and (weights.ndim > 0)
        weights = weights[mask] if wvec_flag else weights

    ret_1d = np.reshape(ret, (-1,))
    ix_1d = idec*shape[1] + ira

    # Return sum of weights, after applying outlier mask, but before applying 'normalize_sum'.
    return utils.scattered_add(ret_1d, ix_1d, weights, normalize_sum=normalize_sum)
    

def map_from_catalog(shape, wcs, gcat, weights=None, rcat=None, rweights=None, normalized=True, allow_outliers=True):
    r"""Project a 3-d galaxy catalog to a 2-d pixell map.

    Function args:

      - ``shape``, ``wcs``: these args define the pixell coordinate system.

      - ``gcat`` (:class:`~kszx.Catalog`): galaxy catalog.
        Must define columns ``ra_deg``, ``dec_deg``.

      - ``weights`` (1-d numpy array, optional): per-galaxy weights.

      - ``rcat``, ``rweights`` (optional): random catalog to be subtracted.
        The rweights are renormalized so that sum(rweights) = -sum(weights).
    
      - ``normalized`` (boolean): If normalized=True, then the output map includes a
        factor ``1 / (pixel area)``. This normalization best represents a sum of delta
        functions $f(x) = \sum_j w_j \delta^2(\theta - \theta_j)$.

      - ``allow_outliers`` (boolean): If 'allow_outliers' is True, then "out-of-bounds"
        galaxies (i.e. galaxies outside the boundaries of the pixell map) will be silently
        discarded. If 'allow_outliers' is False, then out-of-bounds galaxies will raise an
        exception.

    Returns a ``pixell.enmap.ndmap``.
    """

    assert isinstance(gcat, Catalog)
    assert isinstance(rcat, Catalog) or (rcat is None)

    if (rcat is None) and (rweights is not None):
        raise RuntimeError("kszx.healpix_utils.map_from_catalog(): 'rcat' arg is None, but 'rweights' arg is not None")

    weights = utils.asarray(weights, 'kszx.healpix_utils.map_from_catalog', 'weights', dtype=float, allow_none=True)
    rweights = utils.asarray(rweights, 'kszx.healpix_utils.map_from_catalog', 'rweights', dtype=float, allow_none=True)

    assert (weights is None) or (weights.ndim == 0) or (weights.shape == (gcat.size,))
    assert (rweights is None) or (rweights.ndim == 0) or (rweights.shape == (rcat.size,))

    ret = np.zeros(shape)
    wsum = _accum_catalog(ret, shape, wcs, gcat, weights, allow_outliers)

    if rcat is not None:
        _accum_catalog(ret, shape, wcs, rcat, rweights, allow_outliers, normalize_sum = -wsum)

    if normalized:
        ret /= pixell.enmap.pixsizemap(shape, wcs)
    
    return pixell.enmap.enmap(ret, wcs=wcs)


def eval_map_on_catalog(m, catalog, pad=None, return_mask=False):
    r"""Evaluates pixell map 'm' at (ra,dec) values from catalog, and returns a 1-d array.

    Function arguments:

       - ``m`` (pixell.enmap.ndmap): 2-d map to be evaluated.
    
       - ``catalog`` (:class:`~kszx.Catalog`): galaxy catalog defining (ra, dec) values
         where map is evaluated. Must define columns ``ra_deg``, ``dec_deg``.

       - ``pad`` (boolean): Since pixell maps can be partial-sky, some of the (ra,dec)
         pairs may be "outliers", i.e. outside the map footprint. If ``pad=None`` (the default),
         this raises an exception. If 'pad' is specified (e.g. ``pad=0.0``) then outlier values
         are replaced by 'pad'.

       - ``return_mask`` (boolean): Defines what quantity this function returns.

         If ``return_mask=False`` (the default), returns a 1-d array 'map_vals'.

         If ``return_mask=True``, returns a pair ``(map_vals, mask)``, where 'mask' is
         a 1-d boolean array which is True for non-outliers, False for outliers.
    """

    assert isinstance(m, pixell.enmap.ndmap)
    assert isinstance(catalog, Catalog)

    if return_mask and (pad is None):
        raise RuntimeError("kszx.pixell_utils.eval_map_on_catalog(): if return_mask=True, then 'pad' must be specified")
    if pad is not None:
        pad = float(pad)
    
    # pixell_utils.ang2pix() is defined later in this file.
    # It returns (idec,ira) if allow_outliers is False, or (idec,ira,mask) if allow_outliers is True.

    msg = "kszx.pixell_utils.eval_map_on_catalog(): some points are out-of-bounds, you may want to specify the 'pad' arugment"
    t = ang2pix(m.shape, m.wcs, catalog.ra_deg, catalog.dec_deg, allow_outliers=(pad is not None), outlier_errmsg=msg)
    
    idec, ira = t[0], t[1]
    mask = t[2] if (len(t) > 2) else None
    map_vals = m[idec,ira]

    if pad is not None:
        map_vals = np.where(mask, map_vals, pad)

    return (map_vals, mask) if return_mask else map_vals
    

####################################################################################################


def alm2map(alm, shape, wcs):
    """Applies alm2map spherical transform, returns pixell.enmap.ndmap."""
    
    # FIXME map2alm() will blow up with a cryptic error message if it gets a bad WCS (e.g. zeroed).
    # Should find a general test for WCS badness, and add asserts throughout this file.
    
    alm = np.asarray(alm)
    assert alm.ndim == 1
    assert (alm.dtype == complex) or (alm.dtype == np.complex64)

    m = pixell.enmap.enmap(np.zeros(shape), wcs=wcs)
    pixell.curvedsky.alm2map(alm, m)
    return m
    

def map2alm(m, lmax):
    """Applies map2alm spherical transform, returns alms as 1-d complex numpy array."""
    
    assert isinstance(m, pixell.enmap.ndmap)
    assert lmax >= 0
    return pixell.curvedsky.map2alm(m, lmax=lmax)


####################################################################################################


def ang2pix(shape, wcs, ra_deg, dec_deg, allow_outliers=False, outlier_errmsg=None):
    """Given (ra,dec) values on the sky, returns the corresponding pixel indices.

    **Note:** if you are interested in this function in order to pixelize a Catalog,
    you may want to call :func:`~kszx.pixell_utils.map_from_catalog` instead.

    Function arguments:

      - ``shape``, ``wcs``: these args define the pixell coordinate system.

      - ``ra_deg``, ``dec_deg``: numpy arrays with same shape, containing (ra, dec) values.

      - ``allow_outliers`` (boolean): since pixell maps can be partial-sky, pixel indices
        can be out of bounds. In this case:
    
        If ``allow_outliers=False``, we raise an exception.

        If ``allow_outliers=True``, then we "clip" the returned idec/ira arrays
        so that they are in bounds, and return a boolean mask to indicate which
        values are valid (see below).

      - ``outlier_errmsg`` (string, optional): error message if exception is thrown.

    The return value depends on the value of ``allow_outliers``:

      - If ``allow_outliers=False``, returns a pair ``(idec, ira)``.
    
      - If ``allow_outliers=True``, returns a triple ``(idec, ira, mask)``.

    The ``(idec, ira)`` arrays are integer-valued, and contain pixel indices along
    each of the axes of the 2-d pixell map. The ``mask`` array is boolean-valued.
    
    Roughly equivalent to ``healpy.ang2pix(..., latlon=True)``.
    Note that the argument ordering is (ra, dec), consistent with healpy but not pixell!
    """

    # pixell wants a single 'coords' array, with [dec,ra] converted to radians
    coords = np.stack(np.broadcast_arrays(dec_deg, ra_deg), axis=0)   # makes copy
    coords *= (np.pi/180.)

    # pixell.enmap.sky2pix() returns a single output array, where axis 0 is {idec,ira}.
    icoords = pixell.enmap.sky2pix(
        shape,
        wcs,
        coords,
        safe = True,      # pixell does not document this argument
        corner = False,   # indicates that integer coords correspond to pixel centers (not corners)
        bcheck = (not allow_outliers)   # pixell does not document this argument
    )

    icoords = np.asarray(np.round(icoords), dtype=int)

    icoords_clipped = np.clip(
        icoords,
        a_min = 0,
        a_max = np.reshape(shape, (2,)+(1,)*(icoords.ndim-1)) - 1   # note (-1) at end
    )

    mask = np.logical_and(
        icoords[0] == icoords_clipped[0], 
        icoords[1] == icoords_clipped[1]
    )

    idec, ira = icoords_clipped   # not icoords
    
    if allow_outliers:
        return idec, ira, mask
    elif np.all(mask):
        return idec, ira

    if outlier_errmsg is None:
        outlier_errmsg = 'kszx.pixell_utils.ang2pix(): called with allow_outliers=False, and some points are out-of-bounds'
    raise RuntimeError(outlier_errmsg)


####################################################################################################


def uK_arcmin_from_ivar(ivar, max_uK_arcmin=1.0e6):
    r"""Given an inverse variance map, returns associated noise map (in uK-arcmin).

    Based on Mat's orphics library:
       https://github.com/msyriac/orphics/blob/master/orphics/maps.py

    Function args:
    
      - ``ivar`` (pixell map): inverse variance map, units (uK)^{-2}.
        (For example, the return value from :func:`~kszx.act.read_ivar()`.)

      - ``max_uK_arcmin`` (float): max allowed value in the return map (prevents
        dividing by zero).

    Return value is a pixell map with units uK-arcmin.

    Equivalent to::
    
       ps = ivar.pixsizemap()
       uK_arcmin = sqrt(ps/ivar) * (60*180/np.pi)
       return np.minimum(uK_arcmin, max_uK_arcmin)

    but regulated to avoid dividing by zero.
    """

    assert isinstance(ivar, pixell.enmap.ndmap)
    assert max_uK_arcmin > 0

    a = 60*180/np.pi
    ps = ivar.pixsizemap()
    epsilon = 0.9 * np.min(ps) * (a/max_uK_arcmin)**2
    
    uK_arcmin = a * np.sqrt(ps / np.maximum(ivar,epsilon))
    return np.minimum(uK_arcmin, max_uK_arcmin)


def fkp_from_ivar(ivar, cl0, normalize=True, return_wvar=False):
    r"""Given an inverse variance map, returns associated FKP weighting.

    Function args:

      - ``ivar`` (pixell map): inverse variance map, units (uK)^{-2}.
        (For example, the return value from :func:`~kszx.act.read_ivar()`.)

      - ``cl0`` (float): FKP weight function parameter $C_l^{(0)}$.

          - Intuitively, ``cl0`` = "fiducial signal $C_l$ at wavenumber $l$ of interest".
          - ``cl0=0`` corresponds to inverse noise weighting.
          - Large ``cl0`` corresponds to uniform weighing (but ``cl0=np.inf`` won't work).
          - I usually use ``cl0 = 0.01`` for plotting all-sky CMB temperature maps.
          - For kSZ filtering, ``cl0=3e-5`` is a reasonable choice.

      - ``normalize`` (boolean): if True, then we normalize the weight function 
        so that $\max(W(\theta))=1$.

      - ``return_wvar`` (boolean): if True, then we return $W(\theta) / \mbox{ivar}(\theta)$,
        instead of returning $W(\theta)$.
    
    Returns a pixell map.

    The FKP weighting is defined by:
    
    $$\begin{align}
    W(\theta) &= \frac{1}{C_l^{(0)} + N(\theta)} \\
    N(\theta) &\equiv \frac{\mbox{Pixel area}}{\mbox{ivar}(\theta)} \hspace{1cm} \mbox{"Local" noise power spectrum}
    \end{align}$$

    In implementation, in order to avoid divide-by-zero for ivar=0, we compute $W(\theta)$ equivalently as:
    
    $$W(\theta) = \frac{\mbox{ivar}(\theta)}{(\mbox{pixel area}) + C_l^{(0)} \mbox{ivar}(\theta)}$$
    """
    
    assert isinstance(ivar, pixell.enmap.ndmap)
    assert ivar.ndim == 2
    assert np.all(ivar >= 0.0)
    assert cl0 >= 0.0

    wvar = 1.0 / (ivar.pixsizemap() + cl0 * ivar)
    w = wvar * ivar
    
    wmax = np.max(w)
    assert wmax > 0.0

    ret = wvar if return_wvar else w
    
    if normalize:
        ret /= wmax

    return ret


def sensitivity_curve(ivar, step, n):
    r"""Given an ivar map, computes the "sensitivity function" $S(x)$ at x = [ step, ..., n*step].
    
    We define the "sensitivity function" $S(x)$ to be the sky area (in deg^2)
    with sensitivity >= x, where x has units (uK-armcin)^{-2}. Note that $S(x)$
    is a decreasing function of x. This is useful for making a plot of sky sensitivity.

    Function args:

      - ``ivar`` (pixell map): inverse variance map, units (uK)^{-2}.
        (For example, the return value from :func:`~kszx.act.read_ivar()`.)

      - ``step``: units (uK-arcmin)^{-2}, see above.
        Recommend ``step = 1.0e-5`` for ACT.

      - ``n``: number of returned points, see above.

    Returns 1-d array ``[ S(step), S(2*step), ..., S(n*step) ]``.

    Example usage::

      step = 1.0e-5
      nbins = 3000
      ivar = kszx.act.read_ivar(150, dr=6, download=True)    
    
      x = step * np.arange(nbins)
      Sx = kszx.pixell_utils.sensitivity_curve(ivar,step,nbins)

      plt.plot(x, Sx)
      plt.yscale('log')
      plt.ylim(1, 10**5)
      plt.xlabel(r'Threshold $x$ ($\mu$K-arcmin)$^{-2}$')
      plt.ylabel(r'Sky area with sensitivity $> x$ (deg$^2$)')
    """
    
    assert isinstance(ivar, pixell.enmap.ndmap)
    ps = ivar.pixsizemap() * (180./np.pi)**2  # pixel size in deg^2
    x = (ivar / ps) / 3600.                   # sensitivity in (uK-arcmin)^{-2}
    x = np.array(x/step, dtype=int)
    x = np.minimum(x, n)
    x = np.bincount(x.reshape((-1,)), weights=ps.reshape((-1,)), minlength=n+1)
    return np.cumsum(x[::-1])[-2::-1]



import numpy as np

from . import core

from .Box import Box
from .Catalog import Catalog
from .Cosmology import Cosmology


class SurrogateFactory:
    def __init__(self, box, cosmo, randcat, ngal_mean, ngal_rms, ztrue_col='z', kernel='cubic'):
        r"""Helper class for simulating surrogate fields defined on a random catalog.

        Constructor arguments:

          - ``box`` (:class:`~kszx.Box`): defines pixel size, bounding box size, and location of observer.

          - ``cosmo`` (:class:`~kszx.Cosmology`). Used to simulate linear densities/velocities.

          - ``randcat`` (:class:`~kszx.Catalog`): random catalog, defines survey footprint and redshift
            distribution. The randcat must contain columns ``ra_deg`` and ``dec_deg``.

          - ``surr_ngal_mean`` and ``surr_ngal_rms`` (float): In the surrogate sims, I decided to allow $N_{\rm gal}$
            to vary from one surrogate sim to the next. (The idea is to make the surrogate sims more similar to
            mocks, where $N_{\rm gal}$ varies between mocks. Indeed, I find that allowing $N_{\rm gal}$ to vary
            in the surrogate sims does improve overall agreement with mocks.)

            In each surrogate sim, $N_{\rm gal}$ is a Gaussian random variable with mean/rms given by
            the ``surr_ngal_mean`` and ``surr_ngal_rms`` constructor args. If mocks are available, then 
            one way to get sensible values for these arguments is to use the mean/variance in the mocks.
            As a simple placeholder, you could also take ``surr_ngal_rms=0`` (to disable varying $N_{\rm gal}$
            entirely) or ``surr_ngal_rms = sqrt(surr_ngal_mean)`` (Poisson statistics).

          - ``ztrue_col`` (string): name of the randcat column containing redshifts. (If the randcat
            is photometric, then these should be true redshifts, not observed redshifts.)

          - ``kernel`` (string): either ``'cic'`` or ``'cubic'`` (more options will be defined later).

        Most of the interesting members are computed in ``simulate_surrogate()``, but the constructor does
        computes a few useful members:

          - ``self.D`` (1-d array of length nrand): growth function $D(z)$ evaluated on random
            catalog, normalized to $D=1$ at $z=0$.

          - ``self.faH`` (1-d array of length nrand): parameter combination $f_{rsd}(z) H(z) / (1+z)$,
            evaluated on random catalog.

          - ``self.sigma2`` (scalar): variance of linear density field at $z=0$.
        """
        
        assert isinstance(box, Box)
        assert isinstance(cosmo, Cosmology)
        assert isinstance(randcat, Catalog)
        
        assert 0 <= ngal_rms < 0.2 * ngal_mean
        assert ngal_mean + 3*ngal_rms < randcat.size
        
        self.box = box
        self.cosmo = cosmo
        self.kernel = kernel

        self.ngal_mean = ngal_mean
        self.ngal_rms = ngal_rms
        self.ngal_min = ngal_mean - 3 * ngal_rms
        self.ngal_max = ngal_mean + 3 * ngal_rms
        self.nrand = randcat.size

        ztrue = getattr(randcat, ztrue_col)
        self.xyz_true = randcat.get_xyz(cosmo, ztrue_col)
        
        self.D = cosmo.D(z=ztrue, z0norm=True)
        self.faH = cosmo.frsd(z=ztrue) * cosmo.H(z=ztrue) / (1+ztrue)
        self.sigma2 = self._integrate_kgrid(box, cosmo.Plin_z0(box.get_k()))

    
    def simulate_surrogate(self):
        r"""Simulates linear density/velocity fields on the random catalog.

        Initializes the following members:

          - ``self.ngal`` (integer): includes random scatter, see class docstring.
        
          - ``self.delta`` (1-d array of length nrand): linear density field
            $\delta(x)$ evaluated on random catalog.
        
          - ``self.phi`` (1-d array of length nrand): field
            $\phi(x) = \delta(x) / \alpha(x)$ evaluated on random catalog.
        
          - ``self.vr`` (1-d array of length nrand): radial velocity field
            $v_r(x)$  evaluated on random catalog.
        
          - ``self.M`` (1-d array of length nrand): random array with
            $(N_{rand} - N_{gal})$ zeroes and $N_{gal}$ ones.

        Reminder: non-Gaussian galaxy bias takes the form
        $\delta_g(x) = b_g \delta(x) + 2 f_{NL} \delta_c (b_g-1) \phi(x)$.
        """

        ngal = self.ngal_mean + (self.ngal_rms * np.random.normal())
        ngal = np.clip(ngal, self.ngal_min, self.ngal_max)
        ngal = int(ngal+0.5)  # round
        assert 0 < ngal <= self.nrand

        delta = core.simulate_gaussian_field(self.box, self.cosmo.Plin_z0)
        core.apply_kernel_compensation(self.box, delta, self.kernel)

        # Note that phi = delta/alpha is independent of z (in linear theory)
        phi = core.multiply_kfunc(self.box, delta, lambda k: 1.0/self.cosmo.alpha_z0(k=k), dc=0)
        phi = core.interpolate_points(self.box, phi, self.xyz_true, self.kernel, fft=True)
        
        # vr = (faHD/k) delta, evaluated at xyz_true.
        vr = core.multiply_kfunc(self.box, delta, lambda k: 1.0/k, dc=0)
        vr = core.interpolate_points(self.box, vr, self.xyz_true, self.kernel, fft=True, spin=1)
        vr *= self.faH
        vr *= self.D

        delta = core.interpolate_points(self.box, delta, self.xyz_true, self.kernel, fft=True)
        delta *= self.D
            
        # M = random vector with (nrand-ngal) 0s and (ngal) 1s.
        # (This way of generating M is a little slow, but I don't think there's a faster way to do
        # it in numpy, and it's not currently a bottleneck.)
        M = np.zeros(self.nrand)
        M[:ngal] = 1.0
        M = np.random.permutation(M)

        self.ngal = ngal
        self.delta = delta
        self.phi = phi
        self.vr = vr
        self.M = M

    
    @staticmethod
    def _integrate_kgrid(box, kgrid):
        """Helper method for constructor. Could be moved somewhere more general (e.g. kszx.core)."""

        assert kgrid.shape == box.fourier_space_shape
    
        w = np.full(box.fourier_space_shape[-1], 2.0)
        w[0] = 1.0
        w[-1] = 2.0 if (box.real_space_shape[-1] % 2) else 1.0
        w = np.reshape(w, (1,)*(box.ndim-1) + (-1,))

        # FIXME could be optimized
        ret = np.sum(w * kgrid) / box.box_volume
        return np.real(ret)



import numpy as np
import scipy.linalg
import matplotlib.pyplot as plt

from . import utils

from .Catalog import Catalog
from .Cosmology import Cosmology
from .KszPipe import KszPipeOutdir


class PgvLikelihood:
    def __init__(self, data, surrs, bias_matrix, jeffreys_prior=False):
        r"""Likelihood function for one or more kSZ power spectra of the form $P_{gv}(k)$.

        This class can be used to fit $(f_{NL}, b_v)$ values or run MCMCs.
        Currently limited to using $P_{gv}(k)$ only -- it would be interesting to
        extend the analysis to include $P_{gg}(k)$ and/or $P_{vv}(k)$.
        
        The PgvLikelihood constructor has an abstract syntax, which may not be the
        most convenient. Instead of calling the constructor directly, you may want
        to call :meth:`~kszx.PgvLikelihood.from_pipeline_outdir()` as follows::

           dirname = ...   # directory name containing pipeline outputs
           pout = kszx.KszPipeOutdir(dirname)

           # The 'fields=[[1,0]]' argument selects 90 GHz.
           # See the from_pipeline_outdir() docstring for more examples.
           lik = kszx.PgvLikelihood.from_pipeline_outdir(pout, fields=[[1,0]], nkbins=10)

        The rest of this docstring describes the PgvLikelihood constructor syntax
        (even though, as explained above, you probably don't want to call the constructor
        directly!) First we define:

          - B = number of bias parameters in the MCMC (usually 1)
          - V = number of velocity reconstructions (usually 1)
          - K = number of k-bins
          - S = number of surrogate sims
          - D = V*K = number of components in "data vector".

        Note the distinction between a 90+150 GHz "sum" fit (V=1), and a 90+150 GHz
        "joint" fit (V=2). In the first case, we add the 90+150 bandpowers and construct
        a likelihood based on their sum, and in the second case we do a joint fit to
        both sets of bandpowers (i.e. doubling the size of the data vector and
        covariance matrices).
        
        Then the constructor arguments are as follows:

          - ``data``: $P_{gv}(k)$ "data" bandpowers, an array of shape ``(V,K)``.
        
          - ``surrs``: $P_{gv}(k)$ surrogate bandpowers, an array of shape ``(S,2,2,V,K)``.
            The first length-2 index is an fnl exponent in {0,1}.
            The second length-2 index is a bv exponent in {0,1}.

          - ``bias_matrix``: array of shape $(B,V)$, where $B \le V$. This gives the
            correspondence between bias parameters and velocity reconstruction.
            There are basically 3 cases of interest here:

              - ``bias_matrix = [[1,1]]``: analysis of single velocity reconstruction field (V=1).
        
              - ``bias_matrix = [[1,1]]``: joint analysis of two velocity reconstruction fields (V=2),
                with bias assumed to be the same for both freqs.

              - ``bias_matrix = [[1,0],[0,1]]``: joint analysis of two velocity reconstruction
                fields (V=2), with two independent bias parameters.
        """

        data = np.asarray(data, dtype=float)
        surrs = np.asarray(surrs, dtype=float)
        bias_matrix = np.asarray(bias_matrix, dtype=float)
        
        if data.ndim != 2:
            raise RuntimeError(f'got {data.shape=}, expected (V,K)')
        if (surrs.ndim != 5) or (surrs.shape[1:3] != (2,2)):
            raise RuntimeError(f'got {surrs.shape=}, expected (S,2,2,V,K)')
        if bias_matrix.ndim != 2:
            raise RuntimeError(f'got {bias_matrix.shape=}, expected (B,V)')

        if data.shape != surrs.shape[3:]:
            raise RuntimeError(f'data/surrs have inconsistent shapes (expected (V,K) and (S,2,2,V,K), got {data.shape} and {surr.shape})')
        if data.shape[0] != bias_matrix.shape[1]:
            raise RuntimeError(f'data/bias_matrix have inconsistent shapes (expected (V,K) and (B,V), got {data.shape} and {bias_matrix.shape})')
        if bias_matrix.shape[0] > bias_matrix.shape[1]:
            raise RuntimeError(f'expected B <= V, got (B,V)={bias_matrix.shape}')

        self.B = bias_matrix.shape[0]
        self.V = bias_matrix.shape[1]
        self.K = data.shape[1]
        self.S = surrs.shape[0]
        self.D = self.V * self.K
        
        self.data = data                                 # shape (V,K)
        self.data_vector = np.reshape(data, (self.D,))   # shape (D,)
        self.surrs = surrs
        self.bias_matrix = bias_matrix
        self.jeffreys_prior = jeffreys_prior

        self._init_fast_likelihood()


    @staticmethod
    def from_pipeline_outdir(pout, fields, nkbins, multi_bias=None, jeffreys_prior=None):
        r"""Constructs a PgvLikelhood from a :class:`~kszx.KszPipeOutdir` instance.

        Usually more convenient than calling the PgvLikelihood constructor directly.
        Example usage::
        
           dirname = ...   # directory name containing pipeline outputs
           pout = kszx.KszPipeOutdir(dirname)

           # The 'fields=[[1,0]]' argument selects 90 GHz.
           # See below for more examples.
           lik = kszx.PgvLikelihood.from_pipeline_outdir(pout, fields=[[1,0]], nkbins=10)

        The ``pout`` argument is a :class:`~kszx.KszPipe.KszPipeOutdir` object.
        
        The ``fields`` argument is a V-by-2 matrix. Each row of the matrix selects a linear
        combination of the 90+150 GHz velocity reconstructions. For example:
        
           - ``fields = [[1,0]]`` for 90 GHz analysis
           - ``fields = [[0,1]]`` for 150 GHz analysis
           - ``fields = [[1,1]]`` for (90+150) "sum map" analysis
           - ``fields = [[1,-1]]`` for null (90-150) "null map" analysis
           - ``fields = [[1,0],[0,1]]`` for joint analysis with both (90+150 GHz) sets of bandpowers

        The ``multi_bias`` argument is only needed for a joint analysis (i.e. $V > 1$) and determines
        whether each freq channel has an independent bias parameter (``multi_bias=True``), or whether
        all frequency channels have the same bias (``multi_bias=False``).
        """
        
        assert isinstance(pout, KszPipeOutdir)
        assert 4 <= nkbins <= pout.nkbins
        
        fields = np.array(fields)
        if (fields.ndim != 2) or (fields.shape[0] < 1) or (fields.shape[1] != 2):
            raise RuntimeError(f"PgvLikelihood: expected 'fields' arg to have shape (V,2), got shape {fields.shape}")

        K = nkbins
        S = pout.nsurr
        V = fields.shape[0]

        if (V > 1) and (multi_bias is None):
            raise RuntimeError(f"The 'multi_bias' argument is required if nfields > 1.")

        # PgvLikelihood constructor expects data array of shape (V,K)
        pgv_data = pout.pk_data[0,1:3,:K]      # shape (2,K)
        pgv_data = [ (f[0]*pgv_data[0,:] +f[1]*pgv_data[1,:]) for f in fields ]  # shape (V,K)

        # PgvLikelihood constructor expects surrogate array of shape (S,2,2,V,K).
        pgv_surr = pout.pk_surr[:,0:2,2:6,:K]           # shape (S,2,4,K)
        pgv_surr = np.reshape(pgv_surr, (S,2,2,2,K))    # shape (S,2,2,2,K), length-2 indices are (fnl_exponent, freq, bv_exponent)
        pgv_surr = [ (f[0]*pgv_surr[:,:,0,:,:] + f[1]*pgv_surr[:,:,1,:,:]) for f in fields ]  # shape (V,S,2,2,K)
        pgv_surr = np.transpose(pgv_surr, (1,2,3,0,4))  # shape (S,2,2,V,K)

        bias_matrix = np.identity(V) if multi_bias else np.ones((1,V))
        return PgvLikelihood(pgv_data, pgv_surr, bias_matrix, jeffreys_prior)

    
    def specialize_surrogates(self, fnl, bv, flatten):
        r"""Returns a shape ``(S,V,K)`` array, by "specializing" surrogates to specified $(f_{NL}, b_v)$.
        
        The ``bv`` arugment should be an array of shape ``(B,)`` where $B$ is the number
        of bias parameters in the likelihood. If $B=1$, then ``bv`` can be a scalar.

        Convenient but slow! Used in many places, but not :meth:`~kszx.PgvLikelihood.fast_likelihood()`.
        """

        S, K, V, D = self.S, self.K, self.V, self.D
        
        fnl = float(fnl)        
        b = self._validate_bv(bv)        # shape (B,)
        b = np.dot(b, self.bias_matrix)  # shape (V,)
        b = np.reshape(b, (1,V,1))       # shape (1,V,1)

        # Apply fnl, obtaining shape (S,2,V,K)
        s = self.surrs[:,0,:,:,:] + fnl * self.surrs[:,1,:,:,:]

        # Apply bv, obtaining shape (S,V,K)
        s = s[:,0,:,:] + b * s[:,1,:,:]

        # Return either shape (S,V,K) or shape (S,D), depending on whether flatten=True.
        return np.reshape(s,(S,D)) if flatten else s
                
        
    def slow_mean_and_cov(self, fnl, bv):
        r"""Computes mean and covaraince of surrogates, for specified $(f_{NL}, b_v)$.

        Returns ``(mean, cov)`` where ``mean.shape=(D,)`` and ``cov.shape=(D,D).``

        The ``bv`` arugment should be an array of shape ``(B,)`` where $B$ is the number
        of bias parameters in the likelihood. If $B=1$, then ``bv`` can be a scalar.
        
        Brute force implementation: we call :meth:`kszx.PgvLikelihood.specialize_surrogates()``,
        then compute the mean and covariance with ``np.mean()`` and ``np.cov()``.
        """
        
        s = self.specialize_surrogates(fnl, bv, flatten=True)  # shape (S,D)
        mean = np.mean(s, axis=0)       # shape (D,)
        cov = np.cov(s, rowvar=False)   # shape (D, D)
        return mean, cov

    
    def slow_mean_and_cov_gradients(self, fnl, bv):
        r"""Computes the gradient of the mean and covariance with respect to $(f_{NL}, b_v)$.

        Returns ``grad_mean, grad_cov``, where both gradients are represented as arrays
        with an extra length-(B+1) axis, i.e.

          - ``mean.shape = (D,)  =>  grad_mean.shape = (B+1,D)``
          - ``cov.shape = (D,D)  =>  grad_cov.shape = (B+1,D,D)``
          
        Uses boneheaded algorithm: since mean, cov are at most quadratic in $f_{NL}$ and $b_v$,
        naive finite difference is exact (and independent of step sizes).

        The ``bv`` arugment should be an array of shape ``(B,)`` where $B$ is the number
        of bias parameters in the likelihood. If $B=1$, then ``bv`` can be a scalar.

        This method is only used to compute log-likelihoods with the Jeffreys prior.
        """

        B, D = self.B, self.D

        fnl = float(fnl)        
        bv = self._validate_bv(bv)   # shape (B,)

        # Parameter vectors of shape (B+1,)
        x0 = np.concatenate(((fnl,), bv))
        dx = np.concatenate(((50,), np.full(B,0.1)))
        
        mu_p = np.zeros((B+1, D))
        mu_n = np.zeros((B+1, D))
        cov_p = np.zeros((B+1, D, D))
        cov_n = np.zeros((B+1, D, D))

        for i in range(B+1):
            xp = np.copy(x0)
            xn = np.copy(x0)
            xp[i] += dx[i]
            xn[i] -= dx[i]

            mu_p[i,:], cov_p[i,:,:] = self.slow_mean_and_cov(xp[0], xp[1:])
            mu_n[i,:], cov_n[i,:,:] = self.slow_mean_and_cov(xn[0], xn[1:])
   
        grad_mu = (mu_p - mu_n) / (2 * dx.reshape((B+1,1)))
        grad_cov = (cov_p - cov_n) / (2 * dx.reshape((B+1,1,1)))
 
        return grad_mu, grad_cov
        
        
    def slow_log_likelihood(self, fnl, bv):
        r"""Returns the log-likelihood at the specified $(f_{NL}, b_v)$.

        The ``bv`` arugment should be an array of shape ``(B,)`` where $B$ is the number
        of bias parameters in the likelihood. If $B=1$, then ``bv`` can be a scalar.
        """
        
        mean, cov = self.slow_mean_and_cov(fnl, bv)

        x = self.data_vector - mean
        cinv = np.linalg.inv(cov)
        sign, logabsdet = np.linalg.slogdet(cov)
        assert sign == 1

        # log L = -(1/2) log(det C) - (1/2) x^T C^{-1} x
        logL = -0.5 * np.dot(x, np.dot(cinv,x))
        logL -= 0.5 * logabsdet

        if self.jeffreys_prior:
            grad_mu, grad_cov = self.slow_mean_and_cov_gradients(fnl, bv)
            cinv_dmu = [ np.dot(cinv,dmu) for dmu in grad_mu ]
            cinv_dc = [ np.dot(cinv,dc) for dc in grad_cov ]

            # (B+1)-by-(B+1) Fisher matrix            
            B = self.B
            f = np.zeros((B+1,B+1))
            for i in range(B+1):
                for j in range(B+1):
                    f[i,j] += np.dot(grad_mu[i], cinv_dmu[j])  # no factor (1/2)
                    f[i,j] += 0.5 * np.trace(np.dot(cinv_dc[i], cinv_dc[j]))

            # Jeffreys prior is equivalent to including sqrt(det(F)) in the likelihood.
            sign, logabsdet_F = np.linalg.slogdet(f)
            logL += 0.5 * logabsdet_F
            assert sign == 1
            
        return logL

    
    def run_mcmc(self, nwalkers=8, nsamples=10000, discard=1000, thin=5):
        r"""Initializes ``self.samples`` to an array of shape (N,B+1) where N is large."""

        import emcee
        print(f'MCMC start: {nwalkers=}, {nsamples=}, {discard=}, {thin=}')

        x0 = np.zeros((nwalkers, self.B+1))
        x0[:,0] = np.random.uniform(-50, 50, size=nwalkers)  # fnl
        x0[:,1:] = np.random.uniform(0.5, 1.0, size=(nwalkers,self.B))  # bv

        logL = lambda x: self.fast_log_likelihood(x[0], x[1:])
        sampler = emcee.EnsembleSampler(nwalkers, self.B+1, logL)
        sampler.run_mcmc(x0, nsamples)
        self.samples = sampler.get_chain(discard=discard, thin=thin, flat=True)
        
        print('MCMC done. To see the results, call the show_mcmc() method.')


    def show_mcmc(self, title=None):
        r"""Makes a corner plot from MCMC results. Intended to be called from jupyter."""

        if not hasattr(self, 'samples'):
            raise RuntimeError('Must call PgvLikelihood.run_mcmc() before PgvLikelihood.show_mcmc().')
        
        import corner
        
        fig = corner.corner(self.samples, bins=100, range=(0.99,0.99), labels=[r'$f_{NL}$',r'$b_v$'])
        if title is not None:
            fig.suptitle(title)
        
        plt.show()

        fnl_samples = self.samples[:,0]  # shape (nsamp,)
        bv_samples = self.samples[:,1:]  # shape (nsamp,B)
        qlevels = [ 0.025, 0.16, 0.5, 0.84, 0.975 ]
        
        fnl_quantiles = np.quantile(fnl_samples, qlevels)
        for q,fnl in zip(qlevels, fnl_quantiles):
            s = f'  ({(fnl-fnl_quantiles[2]):+.03f})' if (q != 0.5) else ''
            print(f'{(100*q):.03f}% quantile: {fnl=:.03f}{s}')

        for b in range(self.B):
            bv_quantiles = np.quantile(bv_samples[:,b], qlevels)
            print(f'\nBias parameter {b}')
            for q,bv in zip(qlevels, bv_quantiles):
                s = f'  ({(bv-bv_quantiles[2]):+.03f})' if (q != 0.5) else ''
                print(f'{(100*q):.03f}% quantile: {bv=:.03f}{s}')

        print(f'\nSNR: {self.compute_snr():.03f}')


    def analyze_chi2(self, fnl, bv, ddof=None):
        r"""Computes a $\chi^2$ statistic, which compares $P_{gv}^{data}(k)$ to model with given $(f_{NL}, b_v)$.

        Returns ($\chi^2$, $N_{dof}$, $p$-value).

        The ``bv`` arugment should be an array of shape ``(B,)`` where $B$ is the number
        of bias parameters in the likelihood. If $B=1$, then ``bv`` can be a scalar.
        
        The ``ddof`` argument is used to compute the number of degrees of freedom, as
        ``ndof = nkbins - ddof``. If ``ddof=None``, then it will be equal to the
        number of nonzero (fnl, bias) params. (This is usually the correct choice.)
        """
        
        fnl = float(fnl)        
        bv = self._validate_bv(bv)   # shape (B,)
        mean, cov = self.slow_mean_and_cov(fnl, bv)

        if ddof is None:
            ddof = 1 if (fnl != 0) else 0
            ddof += np.count_nonzero(bv)
        
        x = self.data_vector - mean
        chi2 = np.dot(x, np.linalg.solve(cov,x))
        ndof = self.K - ddof
        pte = scipy.stats.chi2.sf(chi2, ndof)
        
        return chi2, ndof, pte
        
    
    def fit_fnl_and_bv(self, fnl0=0, bv0=0.3):
        r"""Returns $(f_{NL}, b_v)$ obtained by maximizing joint likelihood.
        
        Note that $b_v$ is represented as a 1-d array with shape ``(B,)``, even if $B=1$.
        """

        x0 = np.zeros(self.B+1)
        x0[0] = fnl0
        x0[1:] = bv0

        f = lambda x: -self.fast_log_likelihood(x[0], x[1:])  # note minus sign
        result = scipy.optimize.minimize(f, x0, method='Nelder-Mead')
        
        fnl = result.x[0]
        bv = result.x[1:]
        return fnl, bv        

        
    def fit_bv(self, fnl=0, bv0=0.3):
        r"""Returns $b_v$ obtained by maximizing conditional likelihood at the given $f_{NL}.
        
        Note that $b_v$ is represented as a 1-d array with shape ``(B,)``, even if $B=1$.
        """

        x0 = np.full(self.B, bv0)
        f = lambda x: -self.fast_log_likelihood(fnl, x)  # note minus sign
        result = scipy.optimize.minimize(f, x0, method='Nelder-Mead')
        
        bv = result.x
        return bv


    def compute_snr(self):
        r"""Returns total SNR of $P_{gv}(k)$. Does not assume a model for $P_{gv}(k)$."""
        
        # This implementation works even if there are multiple bias params (B > 1).
        B, D = self.B, self.D
        
        _, cov = self.slow_mean_and_cov(0, np.zeros(self.B))                # discard mean
        grad_mu, _ = self.slow_mean_and_cov_gradients(0, np.zeros(self.B))  # discard grad_cov
        m = grad_mu[1:,:]           # shape (B,D)       
        d = self.data_vector        # shape (D,)

        cinv_m = np.linalg.solve(cov, m.T)  # shape (D,B)
        h = np.dot(m, cinv_m)               # shape (B,B)
        g = np.dot(d, cinv_m)

        dchisq = np.dot(g, np.linalg.solve(h, g))
        return np.sqrt(dchisq)
        

    def _validate_bv(self, bv):
        """Helper method: converts 'bv' argument to a 1-d array of length B, and returns it."""

        bv = np.asarray(bv, dtype=float)       
        
        if (bv.ndim == 0) and (self.B == 1):
            return np.reshape(bv, (1,))
            
        if bv.shape != (self.B,):
            raise RuntimeError(f'Got {bv.shape=}, expected 1-d array of length {B=}')

        return bv

    
    ####################################################################################################
    #
    # "Fast" likelihood starts here.
    #
    # This code is completely unreadable, but there are unit tests which verify that the fast_*
    # functions are equivalent o their slow_* equivalents.


    def _init_fast_likelihood(self):
        S, V, K, D = self.S, self.V, self.K, self.D
        
        # xmu = shape (2,2,V,K), reshaped to (2,2*D).
        # Length-2 axes are fnl exponent and bv exponent.
        
        t = np.mean(self.surrs, axis=0)      # shape (2,2,D)
        self.xmu = np.reshape(t, (2,2*D))
        
        # xcov = shape (3,2,V,K,2,V,K), reshaped to (3,4*D*D)
        # Length-3 axis is fnl exponent {0,1,2}, and length-2 axes are bv exponents {0,1}.

        t = np.reshape(self.surrs, (S,4*D))      # shape (S, 4D)
        t = np.cov(t, rowvar=False)              # shape (4D,4D)
        t = np.reshape(t, (2,2*D,2,2*D))         # shape (2,2D,2,2D) where length-2 axes are fnl exponents
        t = np.array([ t[0,:,0,:], t[0,:,1,:]+t[1,:,0,:], t[1,:,1,:] ])   # shape (3,2D,2D)
        t = np.reshape(t, (3,4*D*D))         # shape (3,4D^2)
        self.xcov = np.copy(t)               # make contiguous


    def fast_mean_and_cov(self, fnl, bv, grad=False):
        r"""Equivalent to :meth:`~kszx.PgvLikelihood.slow_mean_and_cov()`, but faster. Intended for use in MCMC.
        
        Note that bv must be a 1-d array of length B, i.e. scalar is not allowed."""
        
        B, D, V, K = self.B, self.D, self.V, self.K
        f3 = np.array((1.0, fnl, fnl**2))

        bv = np.dot(bv, self.bias_matrix)  # shape (V,)
        bv20 = np.reshape(bv, (V,1))
        bv42 = np.reshape(bv, (1,1,V,1))
        bv50 = np.reshape(bv, (V,1,1,1,1))

        # Reminder: xmu = shape (2,2,V,K), reshaped to (2,2*D).
        mu0 = np.dot(f3[:2], self.xmu)       # shape (2*D)
        mu0 = np.reshape(mu0, (2,V,K))       # shape (2,V,K)
        mu = mu0[0,:,:] + bv20 * mu0[1,:,:]  # shape (V,K)
        mu = np.reshape(mu, (D,))

        # Reminder: xcov = shape (3,2,V,K,2,V,K), reshaped to (3,4*D*D)
        cov0 = np.dot(f3, self.xcov)                         # shape (4*D*D)
        cov0 = np.reshape(cov0, (2,V,K,2,V,K))               # shape (2,V,K,2,V,K)
        cov1 = cov0[0,:,:,:,:,:] + bv50 * cov0[1,:,:,:,:,:]  # shape (V,K,2,V,K)
        cov = cov1[:,:,0,:,:] + bv42 * cov1[:,:,1,:,:]       # shape (V,K,V,K)
        cov = np.reshape(cov, (D,D))

        if not grad:
            return mu, cov

        dmu_dfnl = np.reshape(self.xmu[1,:], (2,V,K))        # shape (2,V,K)
        dmu_dfnl = dmu_dfnl[0,:,:] + bv20 * dmu_dfnl[1,:,:]  # shape (V,K)
        dmu_dbv = mu0[1,:,:]                                 # shape (V,K)

        f2 = np.array((1.0, 2*fnl))
        dcov_dfnl = np.dot(f2, self.xcov[1:])                               # shape (4*D*D)
        dcov_dfnl = np.reshape(dcov_dfnl, (2,V,K,2,V,K))                    # shape (2,V,K,2,V,K)
        dcov_dfnl = dcov_dfnl[0,:,:,:,:,:] + bv50 * dcov_dfnl[1,:,:,:,:,:]  # shape (V,K,2,V,K)
        dcov_dfnl = dcov_dfnl[:,:,0,:,:] + bv42 * dcov_dfnl[:,:,1,:,:]      # shape (V,K,V,K)
        dcov_dbv0 = cov0[1,:,:,:,:,:]                                       # shape (V,K,2,V,K)
        dcov_dbv0 = dcov_dbv0[:,:,0,:,:] + bv42 * dcov_dbv0[:,:,1,:,:]      # shape (V,K,V,K)
        dcov_dbv1 = cov1[:,:,1,:,:]                                         # shape (V,K,V,K)
        
        mu_grad = np.zeros((B+1,V,K))     
        mu_grad[0,:,:] = dmu_dfnl
        mu_grad[1:,:,:] = self.bias_matrix.reshape((B,V,1)) * dmu_dbv.reshape((1,V,K))
        mu_grad = np.reshape(mu_grad, (B+1,D))
        
        cov_grad = np.zeros((B+1,V,K,V,K))
        cov_grad[0,:,:,:,:] = dcov_dfnl
        cov_grad[1:,:,:,:,:] = self.bias_matrix.reshape((B,V,1,1,1)) * dcov_dbv0.reshape((1,V,K,V,K))
        cov_grad[1:,:,:,:,:] += self.bias_matrix.reshape((B,1,1,V,1)) * dcov_dbv1.reshape((1,V,K,V,K))
        cov_grad = np.reshape(cov_grad, (B+1,D,D))
        
        return mu, cov, mu_grad, cov_grad


    def fast_log_likelihood(self, fnl, bv):
        r"""Equivalent to :meth:`~kszx.PgvLikelihood.slow_log_likelihood()`, but faster. Intended for use in MCMC.
        
        Note that ``bv`` must be a 1-d array of length B, i.e. scalar is not allowed."""
        
        if self.jeffreys_prior:
            # Need gradients
            mean, cov, grad_mean, grad_cov = self.fast_mean_and_cov(fnl, bv, grad=True)
        else:
            # No gradients needed
            mean, cov = self.fast_mean_and_cov(fnl, bv, grad=False)
        
        x = self.data_vector - mean
        l = np.linalg.cholesky(cov)
        linv_x = scipy.linalg.solve_triangular(l, x, lower=True)

        # log L = -(1/2) log(det C) - (1/2) x^T C^{-1} x
        logL = -0.5 * np.dot(linv_x, linv_x)
        logL -= np.sum(np.log(l.diagonal()))

        if self.jeffreys_prior:
            B, D = (self.B, self.D)
            linv_dmu = scipy.linalg.solve_triangular(l, grad_mean.T, lower=True)  # shape (D,2)
            f = np.dot(linv_dmu.T, linv_dmu)   # first term in 2-by-2 Fisher matrix

            # Second term in 2-by-2 Fisher matrix
            # F_{ij} = (1/2) Tr(C^{-1} dC_i C^{-1} dC_j)
            #        = (1/2) Tr(S_i S_j)  where S_i = L^{-1} dC_i L^{-T}
            
            t = grad_cov.reshape(((B+1)*D, D))
            u = scipy.linalg.solve_triangular(l, t.T, lower=True)  # shape (D, (B+1)*D)
            u = u.reshape((D*(B+1), D))
            v = scipy.linalg.solve_triangular(l, u.T, lower=True)  # shape (D,D*(B+1))
            v = v.reshape((D*D, B+1))
            f += 0.5 * np.dot(v.T, v)

            # Jeffreys prior is equivalent to including sqrt(det(F)) in the likelihood.
            sign, logabsdet_F = np.linalg.slogdet(f)
            logL += 0.5 * logabsdet_F
            assert sign == 1

        return logL

    
    ############################################  Testing  #############################################
    

    def test_fast_mean_and_cov(self):
        r"""Test fast_mean_and_cov(), by checking that it agrees with slow_mean_and_cov() at 10 random points."""

        for _ in range(10):
            fnl = np.random.uniform(-50, 50)
            bv = np.random.uniform(0, 1, size=(self.B,))
            
            slow_mean, slow_cov = self.slow_mean_and_cov(fnl, bv)
            slow_mean_grad, slow_cov_grad = self.slow_mean_and_cov_gradients(fnl, bv)
            fast_mean, fast_cov, fast_mean_grad, fast_cov_grad = self.fast_mean_and_cov(fnl, bv, grad=True)

            assert np.all(np.abs(slow_mean - fast_mean) < 1.0e-10)
            assert np.all(np.abs(slow_cov - fast_cov) < 1.0e-10)
            assert np.all(np.abs(slow_mean_grad - fast_mean_grad) < 1.0e-10)
            assert np.all(np.abs(slow_cov_grad - fast_cov_grad) < 1.0e-10)
    
    
    def test_fast_likelihood(self):
        r"""Test fast_log_likelihood(), by checking that it agrees with slow_log_likelihood() at 10 random points."""
        
        for _ in range(10):
            fnl = np.random.uniform(-50, 50)
            bv = np.random.uniform(0, 1, size=(self.B,))
            logL_slow = self.slow_log_likelihood(fnl, bv)
            logL_fast = self.fast_log_likelihood(fnl, bv)
            assert np.abs(logL_slow - logL_fast) < 1.0e-10
        
    
    @staticmethod
    def make_random():
        r"""Construct and return a PgvLikelihood with random (data, surrs, bias_matrix).
        
        Useful for standalone testing of 'class PgvLikelihood', in order to construct
        an "interesting" PgvLikelihood, in a situation where KSZ pipeline outputs are
        not available."""

        B = np.random.randint(1, 4)       # number of bias parameters in the MCMC
        V = np.random.randint(B, B+3)     # number of velocity reconstructions
        K = np.random.randint(5, 15)      # number of k-bins
        S = np.random.randint(100, 200)   # number of surrogate sims
        jeffreys_prior = (np.random.uniform() < 0.5)   # boolean

        data = np.random.normal(size=(V,K))
        surrs = np.random.normal(size=(S,2,2,V,K))

        # Randomly generate the bias_matrix (shape (B,V), where B <= V)
        # This is not so straightforward, since we want to avoid small SVD eigenvalues
        # for numerical stability.

        rot1 = utils.random_rotation_matrix(B)
        rot2 = utils.random_rotation_matrix(V)
        svds = np.random.uniform(1.0, 2.0, size=B)
        bias_matrix = np.dot(rot1, svds.reshape((B,1)) * rot2[:B,:])

        return PgvLikelihood(data, surrs, bias_matrix, jeffreys_prior)

        
    @staticmethod
    def run_tests():
        """Runs standalone tests of 'class PgvLikelihood'.
        (Where "standalone" means that no KSZ pipeline outputs are needed.)"""
        
        for _ in range(20):
            lik = PgvLikelihood.make_random()
            lik.test_fast_mean_and_cov()
            lik.test_fast_likelihood()

        print('PgvLikelihod.run_tests(): pass')
